import math

from collections import defaultdict
from typing import Union

import torch
from scvi.module import Classifier
from torch import nn
from torch.optim.lr_scheduler import StepLR

from scvi.train import TrainingPlan

import numpy as np
from torchmetrics.functional import accuracy

from ._module import CPAModule
from ._utils import CPA_REGISTRY_KEYS, FocalLoss
from typing import Optional


class CPATrainingPlan(TrainingPlan):
    def __init__(
            self,
            module: CPAModule,
            covars_to_ncovars: dict,
            n_adv_perts: int,
            lr=5e-4,
            wd=1e-6,
            n_steps_pretrain_ae: Optional[int] = None,
            n_epochs_pretrain_ae: Optional[int] = None,
            n_steps_kl_warmup: Optional[int] = None,
            n_epochs_kl_warmup: Optional[int] = None,
            n_steps_adv_warmup: Optional[int] = None,
            n_epochs_adv_warmup: Optional[int] = None,
            n_epochs_mixup_warmup: Optional[int] = None,
            n_epochs_verbose: Optional[int] = 10,
            mixup_alpha: float = 0.0,
            adv_steps: int = 3,
            reg_adv: float = 1.,
            pen_adv: float = 1.,
            n_hidden_adv: int = 64,
            n_layers_adv: int = 3,
            use_batch_norm_adv: bool = True,
            use_layer_norm_adv: bool = False,
            dropout_rate_adv: float = 0.1,
            adv_lr=3e-4,
            adv_wd=4e-7,
            doser_lr=3e-4,
            doser_wd=4e-7,
            step_size_lr: Optional[int] = 45,
            do_clip_grad: Optional[bool] = False,
            gradient_clip_value: Optional[float] = 3.0,
            drug_weights: Optional[list] = None,
            adv_loss: Optional[str] = 'cce',
    ):
        """
        Training plan for the CPA model. 

        Parameters
        ----------
        module: CPAModule
            The model to train.
        covars_to_ncovars: dict
            Dictionary of covariates to their unique values.
        n_adv_perts: int
            Number of all possible perturbations (single/combinatorial).
        lr: float
            Learning rate for the autoencoder.
        wd: float
            Weight decay for the autoencoder.
        n_steps_pretrain_ae: Optional[int]
            Number of steps to pretrain the autoencoder.
        n_epochs_pretrain_ae: Optional[int]
            Number of epochs to pretrain the autoencoder.
        n_steps_kl_warmup: Optional[int]
            Number of steps to warmup the KL term. Will be effective if `variational` has been set True for the model.
        n_epochs_kl_warmup: Optional[int]
            Number of epochs to warmup the KL term. Will be effective if `variational` has been set True for the model.
        n_steps_adv_warmup: Optional[int]
            Number of steps to warmup the adversarial term.
        n_epochs_adv_warmup: Optional[int]
            Number of epochs to warmup the adversarial term.
        n_epochs_mixup_warmup: Optional[int]
            Number of epochs to warmup the mixup term.
        n_epochs_verbose: Optional[int]
            Number of epochs to print the training progress.
        mixup_alpha: float
            Alpha parameter for the mixup term. set this to 0.0 to disable mixup.
        adv_steps: int
            Number of steps to train the adversarial term.
        reg_adv: float
            Regularization parameter for the adversarial term.
        pen_adv: float
            Penalty parameter for the adversarial term.
        n_hidden_adv: int
            Number of hidden units for the adversarial classifier.
        n_layers_adv: int
            Number of layers for the adversarial classifier.
        use_batch_norm_adv: bool
            Whether to use batch normalization for the adversarial classifier.
        use_layer_norm_adv: bool
            Whether to use layer normalization for the adversarial classifier.
        dropout_rate_adv: float
            Dropout rate for the adversarial classifier.
        adv_lr: float
            Learning rate for the adversarial classifier.
        adv_wd: float
            Weight decay for the adversarial classifier.
        doser_lr: float
            Learning rate for the dosers.
        doser_wd: float
            Weight decay for the dosers.
        step_size_lr: Optional[int]
            Step size for the learning rate scheduler.
        do_clip_grad: Optional[bool]
            Whether to clip the gradients.
        gradient_clip_value: Optional[float]
            Value to clip the gradients. Will be effective if `do_clip_grad` is True.
        drug_weights: Optional[list]
            Weights for the perturbations to be used in the adversarial loss.
        adv_loss: Optional[str]
            Adversarial loss to be used. Can be either 'cce' or 'focal'.
        """
        super().__init__(
            module=module,
            lr=lr,
            weight_decay=wd,
            n_steps_kl_warmup=n_steps_kl_warmup,
            n_epochs_kl_warmup=n_epochs_kl_warmup,
            reduce_lr_on_plateau=False,
            lr_factor=None,
            lr_patience=None,
            lr_threshold=None,
            lr_scheduler_metric=None,
            lr_min=None,
        )

        self.automatic_optimization = False

        self.wd = wd

        self.covars_encoder = covars_to_ncovars

        self.mixup_alpha = mixup_alpha
        self.n_epochs_mixup_warmup = n_epochs_mixup_warmup

        self.n_steps_pretrain_ae = n_steps_pretrain_ae
        self.n_epochs_pretrain_ae = n_epochs_pretrain_ae

        self.n_steps_adv_warmup = n_steps_adv_warmup
        self.n_epochs_adv_warmup = n_epochs_adv_warmup

        self.n_epochs_verbose = n_epochs_verbose

        self.adv_steps = adv_steps

        self.reg_adv = reg_adv
        self.pen_adv = pen_adv

        self.adv_lr = adv_lr
        self.adv_wd = adv_wd

        self.doser_lr = doser_lr
        self.doser_wd = doser_wd

        self.step_size_lr = step_size_lr

        self.do_clip_grad = do_clip_grad
        self.gradient_clip_value = gradient_clip_value

        self.metrics = ['recon_loss', 'KL',
                        'disnt_basal', 'disnt_after',
                        'r2_mean', 'r2_var',
                        'adv_loss', 'penalty_adv', 'adv_perts', 'acc_perts', 'penalty_perts']

        self.epoch_history = defaultdict(list)
        self.n_adv_perts = n_adv_perts

        self.perturbation_classifier = Classifier(
            n_input=self.module.n_latent,
            n_labels=n_adv_perts,
            n_hidden=n_hidden_adv,
            n_layers=n_layers_adv,
            use_batch_norm=use_batch_norm_adv,
            use_layer_norm=use_layer_norm_adv,
            dropout_rate=dropout_rate_adv,
            activation_fn=nn.ReLU,
            logits=True,
        )

        self.covars_classifiers = nn.ModuleDict(
            {
                key: Classifier(n_input=self.module.n_latent,
                                n_labels=len(unique_covars),
                                n_hidden=n_hidden_adv,
                                n_layers=n_layers_adv,
                                use_batch_norm=use_batch_norm_adv,
                                use_layer_norm=use_layer_norm_adv,
                                dropout_rate=dropout_rate_adv,
                                logits=True)
                if len(unique_covars) > 1 else None

                for key, unique_covars in self.covars_encoder.items()
            }
        )

        self.drug_weights = torch.tensor(drug_weights).to(self.device) if drug_weights else torch.ones(
            self.n_adv_perts).to(self.device)

        self.adv_loss = adv_loss.lower()
        self.gamma = 2.0
        if self.adv_loss == 'cce':
            self.adv_loss_drugs = nn.CrossEntropyLoss(weight=self.drug_weights)
            self.adv_loss_fn = nn.CrossEntropyLoss()
        elif self.adv_loss == 'focal':
            self.adv_loss_drugs = FocalLoss(alpha=self.drug_weights, gamma=self.gamma, reduction='mean')
            self.adv_loss_fn = FocalLoss(gamma=self.gamma, reduction='mean')
        
    @property
    def adv_lambda(self):
        slope = self.reg_adv
        if self.n_steps_adv_warmup:
            global_step = self.global_step

            if self.n_steps_pretrain_ae:
                 global_step -= self.n_steps_pretrain_ae

            if global_step <= self.n_steps_adv_warmup:
                proportion = global_step / self.n_steps_adv_warmup
                return slope * proportion
            else:
                return slope
        elif self.n_epochs_adv_warmup is not None:
            current_epoch = self.current_epoch

            if self.n_epochs_pretrain_ae:
                current_epoch -= self.n_epochs_pretrain_ae

            if current_epoch <= self.n_epochs_adv_warmup:
                proportion = current_epoch / self.n_epochs_adv_warmup
                return slope * proportion
            else:
                return slope
        else:
            return slope

    @property
    def alpha_mixup(self):
        if self.n_epochs_mixup_warmup:
            current_epoch = self.current_epoch

            if self.n_epochs_pretrain_ae:
                current_epoch -= self.current_epoch

            if current_epoch <= self.n_epochs_mixup_warmup:
                proportion = current_epoch / self.n_epochs_mixup_warmup

                return self.mixup_alpha * proportion
            else:
                return self.mixup_alpha
        else:
            return self.mixup_alpha

    @property
    def do_start_adv_training(self):
        if self.n_steps_pretrain_ae:
            return self.global_step > self.n_steps_pretrain_ae
        elif self.n_epochs_pretrain_ae:
            return self.current_epoch > self.n_epochs_pretrain_ae
        else:
            return True

    def adversarial_loss(self, tensors, z_basal, mixup_lambda: float = 1.0, compute_penalty=True):
        """Computes adversarial classification losses and regularizations"""
        if compute_penalty:
            z_basal = z_basal.requires_grad_(True)

        covars_dict = dict()
        for covar, unique_covars in self.covars_encoder.items():
            encoded_covars = tensors[covar].view(-1, )  # (batch_size,)
            covars_dict[covar] = encoded_covars

        covars_pred = {}
        for covar in self.covars_encoder.keys():
            if self.covars_classifiers[covar] is not None:
                covar_pred = self.covars_classifiers[covar](z_basal)
                covars_pred[covar] = covar_pred
            else:
                covars_pred[covar] = None

        adv_results = {}

        # Classification losses for different covariates
        for covar, covars in self.covars_encoder.items():
            adv_results[f'adv_{covar}'] = mixup_lambda * self.adv_loss_fn(
                covars_pred[covar],
                covars_dict[covar].long(),
            ) if covars_pred[covar] is not None else torch.as_tensor(0.0).to(self.device) + (
                    1. - mixup_lambda) * self.adv_loss_fn(
                covars_pred[covar],
                covars_dict[covar + '_mixup'].long(),
            ) if covars_pred[covar] is not None else torch.as_tensor(0.0).to(self.device)
            adv_results[f'acc_{covar}'] = accuracy(
                covars_pred[covar].argmax(1), covars_dict[covar].long(), task='multiclass',
                num_classes=len(covars)) \
                if covars_pred[covar] is not None else torch.as_tensor(0.0).to(self.device)

        adv_results['adv_loss'] = sum([adv_results[f'adv_{key}'] for key in self.covars_encoder.keys()])

        perturbations = tensors[CPA_REGISTRY_KEYS.PERTURBATION_KEY].view(-1, )
        perturbations_mixup = tensors[CPA_REGISTRY_KEYS.PERTURBATION_KEY + '_mixup'].view(-1, )

        perturbations_pred = self.perturbation_classifier(z_basal)

        adv_results['adv_perts'] = mixup_lambda * self.adv_loss_drugs(perturbations_pred,
                                                                      perturbations.long()) + (
                                           1. - mixup_lambda) * self.adv_loss_drugs(perturbations_pred,
                                                                                    perturbations_mixup.long())

        adv_results['acc_perts'] = mixup_lambda * accuracy(
            perturbations_pred.argmax(1), perturbations.long().view(-1, ), average='macro',
            num_classes=self.n_adv_perts, task='multiclass',
        ) + (1. - mixup_lambda) * accuracy(
            perturbations_pred.argmax(1), perturbations_mixup.long().view(-1, ), average='macro',
            num_classes=self.n_adv_perts, task='multiclass',
        )

        adv_results['adv_loss'] += adv_results['adv_perts']

        if compute_penalty:
            # Penalty losses
            for covar in self.covars_encoder.keys():
                adv_results[f'penalty_{covar}'] = (
                    torch.autograd.grad(
                        covars_pred[covar].sum(),
                        z_basal,
                        create_graph=True,
                        retain_graph=True,
                        only_inputs=True,
                    )[0].pow(2).mean()
                ) if covars_pred[covar] is not None else torch.as_tensor(0.0).to(self.device)

            adv_results['penalty_adv'] = sum([adv_results[f'penalty_{covar}'] for covar in self.covars_encoder.keys()])

            adv_results['penalty_perts'] = (
                torch.autograd.grad(
                    perturbations_pred.sum(),
                    z_basal,
                    create_graph=True,
                    retain_graph=True,
                    only_inputs=True,
                )[0].pow(2).mean()
            )

            adv_results['penalty_adv'] += adv_results['penalty_perts']
        else:
            for covar in self.covars_encoder.keys():
                adv_results[f'penalty_{covar}'] = torch.as_tensor(0.0).to(self.device)

            adv_results['penalty_perts'] = torch.as_tensor(0.0).to(self.device)
            adv_results['penalty_adv'] = torch.as_tensor(0.0).to(self.device)

        return adv_results
    
    def configure_optimizers(self):
        ae_params = list(filter(lambda p: p.requires_grad, self.module.encoder.parameters())) + \
                    list(filter(lambda p: p.requires_grad, self.module.decoder.parameters())) + \
                    list(filter(lambda p: p.requires_grad, self.module.pert_network.pert_embedding.parameters())) + \
                    list(filter(lambda p: p.requires_grad, self.module.covars_embeddings.parameters()))

        if self.module.recon_loss in ['zinb', 'nb']:
            ae_params += [self.module.px_r]

        optimizer_autoencoder = torch.optim.Adam(
            ae_params,
            lr=self.lr,
            weight_decay=self.wd)

        scheduler_autoencoder = StepLR(optimizer_autoencoder, step_size=self.step_size_lr, gamma=0.9)

        doser_params = list(filter(lambda p: p.requires_grad, self.module.pert_network.dosers.parameters()))
        optimizer_doser = torch.optim.Adam(
            doser_params, lr=self.doser_lr, weight_decay=self.doser_wd,
        )
        scheduler_doser = StepLR(optimizer_doser, step_size=self.step_size_lr, gamma=0.9)

        adv_params = list(filter(lambda p: p.requires_grad, self.perturbation_classifier.parameters())) + \
                     list(filter(lambda p: p.requires_grad, self.covars_classifiers.parameters()))

        optimizer_adversaries = torch.optim.Adam(
            adv_params,
            lr=self.adv_lr,
            weight_decay=self.adv_wd)
        scheduler_adversaries = StepLR(optimizer_adversaries, step_size=self.step_size_lr, gamma=0.9)

        optimizers = [optimizer_autoencoder, optimizer_doser, optimizer_adversaries]
        schedulers = [scheduler_autoencoder, scheduler_doser, scheduler_adversaries]

        if self.step_size_lr is not None:
            return optimizers, schedulers
        else:
            return optimizers

    def training_step(self, batch, batch_idx):
        opt, opt_doser, opt_adv = self.optimizers()

        mixup_alpha = self.alpha_mixup

        batch, mixup_lambda = self.module.mixup_data(batch, alpha=mixup_alpha)

        inf_outputs, gen_outputs = self.module.forward(batch, compute_loss=False,
                                                       inference_kwargs={
                                                           'mixup_lambda': mixup_lambda,
                                                       })

        recon_loss, kl_loss = self.module.loss(
            tensors=batch,
            inference_outputs=inf_outputs,
            generative_outputs=gen_outputs,
        )

        if self.do_start_adv_training:
            if self.adv_steps is None:
                opt.zero_grad()
                opt_doser.zero_grad()

                z_basal = inf_outputs['z_basal']

                adv_results = self.adversarial_loss(tensors=batch,
                                                    z_basal=z_basal,
                                                    mixup_lambda=mixup_lambda,
                                                    compute_penalty=False)

                loss = recon_loss + self.kl_weight * kl_loss - self.adv_lambda * adv_results['adv_loss']

                self.manual_backward(loss)

                if self.do_clip_grad:
                    self.clip_gradients(opt,
                                        gradient_clip_val=self.gradient_clip_value,
                                        gradient_clip_algorithm="norm")
                    self.clip_gradients(opt_doser,
                                        gradient_clip_val=self.gradient_clip_value,
                                        gradient_clip_algorithm="norm")
                opt.step()
                opt_doser.step()

                opt_adv.zero_grad()

                adv_results = self.adversarial_loss(tensors=batch,
                                                    z_basal=z_basal.detach(),
                                                    mixup_lambda=mixup_lambda,
                                                    compute_penalty=True)

                adv_loss = adv_results['adv_loss'] + self.pen_adv * adv_results['penalty_adv']

                self.manual_backward(adv_loss)

                if self.do_clip_grad:
                    self.clip_gradients(opt_adv,
                                        gradient_clip_val=self.gradient_clip_value,
                                        gradient_clip_algorithm="norm")

                opt_adv.step()

            elif batch_idx % self.adv_steps != 0:
                opt_adv.zero_grad()

                z_basal = inf_outputs['z_basal']

                adv_results = self.adversarial_loss(tensors=batch,
                                                    z_basal=z_basal.detach(),
                                                    mixup_lambda=mixup_lambda,
                                                    compute_penalty=True)

                adv_loss = adv_results['adv_loss'] + self.pen_adv * adv_results['penalty_adv']

                self.manual_backward(adv_loss)

                if self.do_clip_grad:
                    self.clip_gradients(opt_adv,
                                        gradient_clip_val=self.gradient_clip_value,
                                        gradient_clip_algorithm="norm")

                opt_adv.step()

            # Model update
            else:
                opt.zero_grad()
                opt_doser.zero_grad()

                z_basal = inf_outputs['z_basal']

                adv_results = self.adversarial_loss(tensors=batch,
                                                    z_basal=z_basal,
                                                    mixup_lambda=mixup_lambda,
                                                    compute_penalty=False)

                loss = recon_loss + self.kl_weight * kl_loss - self.adv_lambda * adv_results['adv_loss']

                self.manual_backward(loss)

                if self.do_clip_grad:
                    self.clip_gradients(opt,
                                        gradient_clip_val=self.gradient_clip_value,
                                        gradient_clip_algorithm="norm")
                    self.clip_gradients(opt_doser,
                                        gradient_clip_val=self.gradient_clip_value,
                                        gradient_clip_algorithm="norm")
                opt.step()
                opt_doser.step()

        else:
            opt.zero_grad()
            opt_doser.zero_grad()

            z_basal = inf_outputs['z_basal']

            loss = recon_loss + self.kl_weight * kl_loss

            self.manual_backward(loss)

            if self.do_clip_grad:
                self.clip_gradients(opt,
                                    gradient_clip_val=self.gradient_clip_value,
                                    gradient_clip_algorithm="norm")
                self.clip_gradients(opt_doser,
                                    gradient_clip_val=self.gradient_clip_value,
                                    gradient_clip_algorithm="norm")

            opt.step()
            opt_doser.step()

            opt_adv.zero_grad()

            adv_results = self.adversarial_loss(tensors=batch,
                                                z_basal=z_basal.detach(),
                                                mixup_lambda=mixup_lambda,
                                                compute_penalty=True)

            adv_loss = adv_results['adv_loss'] + self.pen_adv * adv_results['penalty_adv']

            self.manual_backward(adv_loss)

            if self.do_clip_grad:
                self.clip_gradients(opt_adv,
                                    gradient_clip_val=self.gradient_clip_value,
                                    gradient_clip_algorithm="norm")

            opt_adv.step()

        r2_mean, r2_var = self.module.r2_metric(batch, inf_outputs, gen_outputs, mode='direct')

        for key, val in adv_results.items():
            adv_results[key] = val.item()

        results = adv_results.copy()
        results.update({'recon_loss': recon_loss.item()})
        results.update({'KL': kl_loss.item()})

        results.update({'r2_mean': r2_mean, 'r2_var': r2_var})
        results.update({'r2_mean_lfc': 0.0, 'r2_var_lfc': 0.0})
        results.update({'cpa_metric': 0.0})
        results.update({'disnt_basal': 0.0, 'disnt_after': 0.0})

        return results

    def training_epoch_end(self, outputs):
        for key in self.metrics:
            if key in ['disnt_basal', 'disnt_after']:
                self.epoch_history[key].append(0.0)
            else:
                self.epoch_history[key].append(np.mean([output[key] for output in outputs if output[key] != 0.0]))

        for covar, unique_covars in self.covars_encoder.items():
            if len(unique_covars) > 1:
                key1, key2, key3 = f'adv_{covar}', f'penalty_{covar}', f'acc_{covar}'
                self.epoch_history[key1].append(np.mean([output[key1] for output in outputs if output[key1] != 0.0]))
                self.epoch_history[key2].append(np.mean([output[key2] for output in outputs if output[key2] != 0.0]))
                self.epoch_history[key3].append(np.mean([output[key3] for output in outputs if output[key3] != 0.0]))

        self.epoch_history['epoch'].append(self.current_epoch)
        self.epoch_history['mode'].append('train')

        self.log("recon", self.epoch_history['recon_loss'][-1], prog_bar=True)
        self.log("r2_mean", self.epoch_history['r2_mean'][-1], prog_bar=True)
        self.log("adv_loss", self.epoch_history['adv_loss'][-1], prog_bar=True)
        self.log("acc_pert", self.epoch_history['acc_perts'][-1], prog_bar=True)
        for covar, nc in self.covars_encoder.items():
            if len(nc) > 1:
                self.log(f'acc_{covar}', self.epoch_history[f'acc_{covar}'][-1], prog_bar=True)

        if self.current_epoch > 1 and self.current_epoch % self.step_size_lr == 0:
            sch, sch_doser, sch_adv = self.lr_schedulers()
            sch.step()
            sch_doser.step()
            sch_adv.step()

    def validation_step(self, batch, batch_idx):
        batch, mixup_lambda = self.module.mixup_data(batch, alpha=0.0)  # No mixup during validation

        inf_outputs, gen_outputs = self.module.forward(batch, compute_loss=False,
                                                       inference_kwargs={
                                                           'mixup_lambda': 1.0,
                                                       })

        recon_loss, kl_loss = self.module.loss(
            tensors=batch,
            inference_outputs=inf_outputs,
            generative_outputs=gen_outputs,
        )

        adv_results = {'adv_loss': 0.0, 'cycle_loss': 0.0, 'penalty_adv': 0.0,
                       'adv_perts': 0.0, 'acc_perts': 0.0, 'penalty_perts': 0.0}
        for covar in self.covars_encoder.keys():
            adv_results[f'adv_{covar}'] = 0.0
            adv_results[f'acc_{covar}'] = 0.0
            adv_results[f'penalty_{covar}'] = 0.0

        r2_mean, r2_var = self.module.r2_metric(batch, inf_outputs, gen_outputs, mode='direct')
        disnt_basal, disnt_after = self.module.disentanglement(batch, inf_outputs, gen_outputs)

        results = adv_results
        results.update({'r2_mean': r2_mean, 'r2_var': r2_var})
        results.update({'disnt_basal': disnt_basal})
        results.update({'disnt_after': disnt_after})
        results.update({'KL': kl_loss.item()})
        results.update({'recon_loss': recon_loss.item()})
        results.update({'cpa_metric': r2_mean + 0.5 * r2_var + math.e ** (disnt_after - disnt_basal)})

        return results

    def validation_epoch_end(self, outputs):
        for key in self.metrics:
            self.epoch_history[key].append(np.mean([output[key] for output in outputs if output[key] != 0.0]))

        for covar, unique_covars in self.covars_encoder.items():
            if len(unique_covars) > 1:
                key1, key2, key3 = f'adv_{covar}', f'penalty_{covar}', f'acc_{covar}'
                self.epoch_history[key1].append(np.mean([output[key1] for output in outputs if output[key1] != 0.0]))
                self.epoch_history[key2].append(np.mean([output[key2] for output in outputs if output[key2] != 0.0]))
                self.epoch_history[key3].append(np.mean([output[key3] for output in outputs if output[key3] != 0.0]))

        self.epoch_history['epoch'].append(self.current_epoch)
        self.epoch_history['mode'].append('valid')

        self.log('val_recon', self.epoch_history['recon_loss'][-1], prog_bar=True)
        self.log('cpa_metric', np.mean([output['cpa_metric'] for output in outputs]), prog_bar=False)
        self.log('disnt_basal', self.epoch_history['disnt_basal'][-1], prog_bar=True)
        self.log('disnt_after', self.epoch_history['disnt_after'][-1], prog_bar=True)
        self.log('val_r2_mean', self.epoch_history['r2_mean'][-1], prog_bar=True)
        self.log('val_r2_var', self.epoch_history['r2_var'][-1], prog_bar=False)
        self.log('val_KL', self.epoch_history['KL'][-1], prog_bar=True)

        if self.current_epoch % self.n_epochs_verbose == self.n_epochs_verbose - 1:
            print(f'\ndisnt_basal = {self.epoch_history["disnt_basal"][-1]}')
            print(f'disnt_after = {self.epoch_history["disnt_after"][-1]}')
            print(f'val_r2_mean = {self.epoch_history["r2_mean"][-1]}')
            print(f'val_r2_var = {self.epoch_history["r2_var"][-1]}')




# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from collections import defaultdict

import matplotlib.font_manager
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from adjustText import adjust_text
from sklearn.decomposition import KernelPCA
from sklearn.metrics.pairwise import cosine_similarity

from scipy import stats, sparse
from sklearn.metrics import r2_score

from . import CPA
from ._api import ComPertAPI

from typing import Optional

FONT_SIZE = 13
font = {'size': FONT_SIZE, 'family': 'monospace'}

matplotlib.rc('font', **font)
matplotlib.rc('ytick', labelsize=FONT_SIZE)
matplotlib.rc('xtick', labelsize=FONT_SIZE)


class CompertVisuals:
    """
    A wrapper for automatic plotting CompPert latent drug and dose-response
    curve. Sets up prefix for all files and default dictionaries for atomic
    perturbations and cell types.
    """

    def __init__(self,
                 compert_api: ComPertAPI,
                 fileprefix=None,
                 perts_palette=None,
                 covars_palette=None,
                 plot_params={'fontsize': 13},
                 ):
        """
        Parameters
        ----------
        moc : CompPertAPI
            Variable from ComPertAPI class.
        fileprefix : str, optional (default: None)
            Prefix (with path) to the filename to save all drug in a
            standartized manner. If None, drug are not saved to file.
        perts_palette : dict (default: None)
            Dictionary of colors for the drug of perturbations. Keys
            correspond to perturbations and values to their colors. If None,
            default dicitonary will be set up.
        сovars_palette : dict (default: None)
            Dictionary of colors for the drug of covariates. Keys
            correspond to covariates and values to their colors. If None,
            default dicitonary will be set up.
        """

        self.fileprefix = fileprefix

        self.perturbation_key = compert_api.perturbation_key
        self.dose_key = compert_api.dose_key
        self.covars_key = compert_api.covars_key
        self.measured_points = compert_api.measured_points

        self.unique_perts = compert_api.unique_perts
        self.unique_covars = compert_api.unique_covars

        if perts_palette is None:
            self.perts_palette = dict(zip(self.unique_perts,
                                          get_palette(len(self.unique_perts))))
        else:
            self.perts_palette = perts_palette

        if covars_palette is None:
            self.covars_palette = {}
            for covar in self.unique_covars.keys():
                self.covars_palette[covar] = dict(
                    zip(self.unique_covars[covar], get_palette(len(self.unique_covars[covar]), palette_name='tab10')))
        else:
            self.covars_palette = covars_palette

        if plot_params['fontsize'] is None:
            self.fontsize = FONT_SIZE
        else:
            self.fontsize = plot_params['fontsize']

    def plot_latent_embeddings(self,
                               emb,
                               titlename='Example',
                               kind='perturbations',
                               palette=None,
                               labels=None,
                               dimred='KernelPCA',
                               filename=None,
                               show_text=True
                               ):
        """
        Parameters
        ----------
        emb : np.array
            Multi-dimensional embedding of perturbations or covariates.
        titlename : str, optional (default: 'Example')
            Title.
        kind : int, optional, optional (default: 'perturbations')
            Specify if this is embedding of perturbations, covariates or some
            other. If it is perturbations or covariates, it will use default
            saved dictionaries for colors.
        palette : dict, optional (default: None)
            If embedding of kind not perturbations or covariates, the user can
            specify color dictionary for the embedding.
        labels : list, optional (default: None)
            Labels for the drug.
        dimred : str, optional (default: 'KernelPCA')
            Dimensionality reduction method for plotting low dimensional
            representations. Options: 'KernelPCA', 'UMAPpre', 'UMAPcos', None.
            If None, uses first 2 dimensions of the embedding.
        filename : str (default: None)
            Name of the file to save the plot. If None, will automatically
            generate name from prefix file.
        """
        if filename is None:
            if self.fileprefix is None:
                filename = None
                file_name_similarity = None
            else:
                filename = f'{self.fileprefix}_emebdding.png'
                file_name_similarity = f'{self.fileprefix}_emebdding_similarity.png'
        else:
            file_name_similarity = filename.split('.')[0] + '_similarity.png'

        if labels is None:
            if kind == 'perturbations':
                palette = self.perts_palette
                labels = self.unique_perts
            else:  # covars kind
                palette = self.covars_palette[kind]
                labels = self.unique_covars[kind]

        if len(emb) < 2:
            print(f'Embedding contains only {len(emb)} vectors. Not enough to plot.')
        else:
            plot_embedding(
                fast_dimred(emb, method=dimred),
                labels,
                show_lines=True,
                show_text=show_text,
                col_dict=palette,
                title=titlename,
                file_name=filename,
                fontsize=self.fontsize
            )

            plot_similarity(
                emb,
                labels,
                col_dict=palette,
                fontsize=self.fontsize,
                file_name=file_name_similarity
            )

    def plot_contvar_response2D(self,
                                df_response2D,
                                df_ref=None,
                                levels=15,
                                figsize=(4, 4),
                                xlims=(0, 1.03),
                                ylims=(0, 1.03),
                                palette="coolwarm",
                                response_name='response',
                                title_name=None,
                                fontsize=None,
                                postfix='',
                                filename=None,
                                alpha=0.4,
                                sizes=(40, 160),
                                logdose=False,
                                file_format='png'):

        """
        Parameters
        ----------
        df_response2D : pd.DataFrame
            Data frame with responses of combinations with columns=(dose1, dose2,
            response).
        levels: int, optional (default: 15)
            Number of levels for contour plot.
        response_name : str (default: 'response')
            Name of column in df_response to plot as response.
        alpha: float (default: 0.4)
            Transparency of the background contour.
        figsize: tuple (default: (4,4))
            Size of the figure in inches.
        palette : dict, optional (default: None)
            Colors dictionary for perturbations to plot.
        title_name : str, optional (default: None)
            Title for the plot.
        postfix : str, optional (defualt: '')
            Postfix to add to the output file name to save the model.
        filename : str, optional (defualt: None)
            Name of the file to save the plot.  If None, will automatically
            generate name from prefix file.
        logdose: bool (default: False)
            If True, dose values will be log10. 0 values will be mapped to
            minumum value -1,e.g.
            if smallest non-zero dose was 0.001, 0 will be mapped to -4.
        """
        sns.set_style("white")

        if (filename is None) and not (self.fileprefix is None):
            filename = f'{self.fileprefix}_{postfix}response2D.png'
        if fontsize is None:
            fontsize = self.fontsize

        x_name, y_name = df_response2D.columns[:2]

        x = df_response2D[x_name].values
        y = df_response2D[y_name].values

        if logdose:
            x = log10_with0(x)
            y = log10_with0(y)

        z = df_response2D[response_name].values

        n = int(np.sqrt(len(x)))

        X = x.reshape(n, n)
        Y = y.reshape(n, n)
        Z = z.reshape(n, n)

        fig, ax = plt.subplots(figsize=figsize)

        CS = ax.contourf(X, Y, Z, cmap=palette, levels=levels, alpha=alpha)
        CS = ax.contour(X, Y, Z, levels=15, cmap=palette)
        ax.clabel(CS, inline=1, fontsize=fontsize)
        ax.set(xlim=(0, 1), ylim=(0, 1))
        ax.axis("equal")
        ax.axis("square")
        ax.yaxis.set_tick_params(labelsize=fontsize)
        ax.xaxis.set_tick_params(labelsize=fontsize)
        ax.set_xlabel(x_name, fontsize=fontsize, fontweight="bold")
        ax.set_ylabel(y_name, fontsize=fontsize, fontweight="bold")
        ax.set_xlim(xlims)
        ax.set_ylim(ylims)

        # sns.despine(left=False, bottom=False, right=True)
        sns.despine()

        if not (df_ref is None):
            sns.scatterplot(
                x=x_name,
                y=y_name,
                hue='split',
                size='num_cells',
                sizes=sizes,
                alpha=1.,
                palette={'train': '#000000', 'training': '#000000', 'ood': '#e41a1c'},
                data=df_ref, ax=ax)
            ax.legend_.remove()

        ax.set_title(title_name, fontweight="bold", fontsize=fontsize)
        plt.tight_layout()

        if filename:
            save_to_file(fig, filename)

    def plot_contvar_response(self,
                              df_response,
                              response_name='response',
                              var_name=None,
                              df_ref=None,
                              palette=None,
                              title_name=None,
                              postfix='',
                              ref_name='origin',
                              xlabelname=None,
                              filename=None,
                              logdose=False,
                              fontsize=None,
                              legend=True,
                              measured_points=None,
                              bbox=(1.35, 1.),
                              figsize=(7., 4.)
                              ):
        """
        Parameters
        ----------
        df_response : pd.DataFrame
            Data frame of responses.
        response_name : str (default: 'response')
            Name of column in df_response to plot as response.
        var_name : str, optional  (default: None)
            Name of conditioning variable, e.g. could correspond to covariates.
        df_ref : pd.DataFrame, optional  (default: None)
            Reference values. Fields for plotting should correspond to
            df_response.
        palette : dict, optional (default: None)
            Colors dictionary for perturbations to plot.
        title_name : str, optional (default: None)
            Title for the plot.
        postfix : str, optional (defualt: '')
            Postfix to add to the output file name to save the model.
        filename : str, optional (defualt: None)
            Name of the file to save the plot.  If None, will automatically
            generate name from prefix file.
        logdose: bool (default: False)
            If True, dose values will be log10. 0 values will be mapped to
            minumum value -1,e.g.
            if smallest non-zero dose was 0.001, 0 will be mapped to -4.
        figsize: tuple (default: (7., 4.))
            Size of output figure
        """
        if (filename is None) and not (self.fileprefix is None):
            filename = f'{self.fileprefix}_{postfix}response.png'

        if fontsize is None:
            fontsize = self.fontsize

        if logdose:
            dose_name = f'log10-{self.dose_key}'
            df_response[dose_name] = log10_with0(df_response[self.dose_key].values)
            if not (df_ref is None):
                df_ref[dose_name] = log10_with0(df_ref[self.dose_key].values)
        else:
            dose_name = self.dose_key

        if var_name is None:
            if len(self.unique_covars) > 1:
                var_name = self.covars_key
            else:
                var_name = self.perturbation_key

        if palette is None:
            if var_name == self.perturbation_key:
                palette = self.perts_palette
            elif var_name == self.covars_key:
                palette = self.covars_palette

        plot_dose_response(df_response,
                           dose_name,
                           var_name,
                           xlabelname=xlabelname,
                           df_ref=df_ref,
                           ref_name=ref_name,
                           response_name=response_name,
                           title_name=title_name,
                           use_ref_response=(not (df_ref is None)),
                           col_dict=palette,
                           plot_vertical=False,
                           f1=figsize[0],
                           f2=figsize[1],
                           fname=filename,
                           logscale=measured_points,
                           measured_points=measured_points,
                           bbox=bbox,
                           legend=legend,
                           fontsize=fontsize,
                           format='png')

    def plot_scatter(
            self,
            df,
            x_axis,
            y_axis,
            hue=None,
            size=None,
            style=None,
            figsize=(4.5, 4.5),
            title=None,
            palette=None,
            filename=None,
            alpha=.75,
            sizes=(30, 90),
            text_dict=None,
            postfix='',
            fontsize=14):

        sns.set_style("white")

        if (filename is None) and not (self.fileprefix is None):
            filename = f'{self.fileprefix}_scatter{postfix}.png'

        if fontsize is None:
            fontsize = self.fontsize

        fig = plt.figure(figsize=figsize)
        ax = plt.gca()
        sns.scatterplot(
            x=x_axis,
            y=y_axis,
            hue=hue,
            style=style,
            size=size,
            sizes=sizes,
            alpha=alpha,
            palette=palette,
            data=df)

        ax.legend_.remove()
        ax.set_xlabel(x_axis, fontsize=fontsize)
        ax.set_ylabel(y_axis, fontsize=fontsize)
        ax.xaxis.set_tick_params(labelsize=fontsize)
        ax.yaxis.set_tick_params(labelsize=fontsize)
        ax.set_title(title)
        if not (text_dict is None):
            texts = []
            for label in text_dict.keys():
                texts.append(
                    ax.text(
                        text_dict[label][0],
                        text_dict[label][1],
                        label,
                        fontsize=fontsize
                    )
                )

            adjust_text(
                texts,
                arrowprops=dict(arrowstyle='-', color='black', lw=0.1),
                ax=ax
            )

        plt.tight_layout()

        if filename:
            save_to_file(fig, filename)


def log10_with0(x):
    mx = np.min(x[x > 0])
    x[x == 0] = mx / 10
    return np.log10(x)


def get_palette(
        n_colors,
        palette_name='Set1'
):
    try:
        palette = sns.color_palette(palette_name)
    except:
        print('Palette not found. Using default palette tab10')
        palette = sns.color_palette()
    while len(palette) < n_colors:
        palette += palette

    return palette


def fast_dimred(emb, method='KernelPCA'):
    """
    Takes high dimensional drug and produces a 2-dimensional representation
    for plotting.
    emb: np.array
        Embeddings matrix.
    method: str (default: 'KernelPCA')
        Method for dimensionality reduction: KernelPCA, UMAPpre, UMAPcos, tSNE.
        If None return first 2 dimensions of the embedding vector.
    """
    if method is None:
        return emb[:, :2]
    elif method == 'KernelPCA':
        similarity_matrix = cosine_similarity(emb)
        np.fill_diagonal(similarity_matrix, 1.0)
        X = KernelPCA(n_components=2, kernel="precomputed") \
            .fit_transform(similarity_matrix)
    else:
        raise NotImplementedError

    return X


def plot_dose_response(df,
                       contvar_key,
                       perturbation_key,
                       df_ref=None,
                       response_name='response',
                       use_ref_response=False,
                       palette=None,
                       col_dict=None,
                       fontsize=8,
                       measured_points=None,
                       interpolate=True,
                       f1=7,
                       f2=3.,
                       bbox=(1.35, 1.),
                       legend=True,
                       ref_name='origin',
                       title_name='None',
                       plot_vertical=True,
                       fname=None,
                       logscale=None,
                       xlabelname=None,
                       format='png'):
    """Plotting decoding of the response with respect to dose.

    Params
    ------
    df : `DataFrame`
        Table with columns=[perturbation_key, contvar_key, response_name].
        The last column is always "response".
    contvar_key : str
        Name of the column in df for values to use for x axis.
    perturbation_key : str
        Name of the column in df for the perturbation or covariate to plot.
    response_name: str (default: response)
        Name of the column in df for values to use for y axis.
    df_ref : `DataFrame` (default: None)
        Table with the same columns as in df to plot ground_truth or another
        condition for comparison. Could
        also be used to just extract reference values for x-axis.
    use_ref_response : bool (default: False)
        A flag indicating if to use values for y axis from df_ref (True) or j
        ust to extract reference values for x-axis.
    col_dict : dictionary (default: None)
        Dictionary with colors for each value in perturbation_key.
    bbox : tuple (default: (1.35, 1.))
        Coordinates to adjust the legend.
    plot_vertical : boolean (default: False)
        Flag if to plot reference values for x axis from df_ref dataframe.
    f1 : float (default: 7.0))
        Width in inches for the plot.
    f2 : float (default: 3.0))
        Hight in inches for the plot.
    fname : str (default: None)
        Name of the file to export the plot. The name comes without format
        extension.
    format : str (default: png)
        Format for the file to export the plot.
    """
    sns.set_style("white")
    if use_ref_response and not (df_ref is None):
        df[ref_name] = 'predictions'
        df_ref[ref_name] = 'observations'
        if interpolate:
            df_plt = pd.concat([df, df_ref])
        else:
            df_plt = df
    else:
        df_plt = df

    atomic_drugs = np.unique(df[perturbation_key].values)

    if palette is None:
        current_palette = get_palette(len(list(atomic_drugs)))

    if col_dict is None:
        col_dict = dict(
            zip(
                list(atomic_drugs),
                current_palette
            )
        )

    fig = plt.figure(figsize=(f1, f2))
    ax = plt.gca()

    if use_ref_response:
        sns.lineplot(
            x=contvar_key,
            y=response_name,
            palette=col_dict,
            hue=perturbation_key,
            style=ref_name,
            dashes=[(1, 0), (2, 1)],
            legend='full',
            style_order=['predictions', 'observations'],
            data=df_plt.reset_index(drop=True), ax=ax)

        df_ref = df_ref.replace('training_treated', 'train')
        sns.scatterplot(
            x=contvar_key,
            y=response_name,
            hue='split',
            size='num_cells',
            sizes=(10, 100),
            alpha=1.,
            palette={'train': '#000000', 'training': '#000000', 'ood': '#e41a1c'},
            data=df_ref, ax=ax)

        if legend:
            ax.legend_.remove()
    else:
        sns.lineplot(x=contvar_key, y=response_name,
                     palette=col_dict,
                     hue=perturbation_key,
                     data=df_plt, ax=ax)
        ax.legend(
            loc='upper right',
            bbox_to_anchor=bbox,
            fontsize=fontsize)

    if not (title_name is None):
        ax.set_title(title_name, fontsize=fontsize, fontweight='bold')
    ax.grid('off')

    if xlabelname is None:
        ax.set_xlabel(contvar_key, fontsize=fontsize)
    else:
        ax.set_xlabel(xlabelname, fontsize=fontsize)

    ax.set_ylabel(f"{response_name}", fontsize=fontsize)

    ax.xaxis.set_tick_params(labelsize=fontsize)
    ax.yaxis.set_tick_params(labelsize=fontsize)

    if not (logscale is None):
        ax.set_xticks(np.log10(logscale))
        ax.set_xticklabels(logscale, rotation=90)

    if not (df_ref is None):
        atomic_drugs = np.unique(df_ref[perturbation_key].values)
        for drug in atomic_drugs:
            if drug in df[perturbation_key].unique():
                x = df_ref[df_ref[perturbation_key] == drug][contvar_key].values
                m1 = np.min(df[df[perturbation_key] == drug][response_name].values)
                m2 = np.max(df[df[perturbation_key] == drug][response_name].values)

                if plot_vertical:
                    for x_dot in x:
                        ax.plot([x_dot, x_dot], [m1, m2], ':', color='black',
                                linewidth=.5, alpha=0.5)

    fig.tight_layout()
    if fname:
        plt.savefig(f'{fname}.{format}', format=format)

    return fig


def plot_uncertainty_comb_dose(
        compert_api,
        cov,
        pert,
        N=11,
        metric='cosine',
        measured_points=None,
        cond_key='condition',
        vmin=None,
        vmax=None,
        sizes=(40, 160),
        df_ref=None,
        xlims=(0, 1.03),
        ylims=(0, 1.03),
        fixed_drugs='',
        fixed_doses='',
        title=True,
        filename=None
):
    """Plotting uncertainty for a single perturbation at a dose range for a
    particular covariate.

    Params
    ------
    compert_api
        Api object for the model class.
    cov : str
        Name of covariate.
    pert : str
        Name of the perturbation.
    N : int
        Number of dose values.
    metric: str (default: 'cosine')
        Metric to evaluate uncertainty.
    measured_points : dict (default: None)
        A dicitionary of dictionaries. Per each covariate a dictionary with
        observed doses per perturbation, e.g. {'covar1': {'pert1':
        [0.1, 0.5, 1.0], 'pert2': [0.3]}
    cond_key : str (default: 'condition')
        Name of the variable to use for plotting.
    filename : str (default: None)
        Full path to the file to export the plot. File extension should be
        included.

    Returns
        -------
        pd.DataFrame of uncertainty estimations.
    """

    df_list = []
    for i in np.round(np.linspace(0, 1, N), decimals=2):
        for j in np.round(np.linspace(0, 1, N), decimals=2):
            df_list.append(
                {
                    'cell_type': cov,
                    'condition': pert + fixed_drugs,
                    'dose_val': str(i) + '+' + str(j) + fixed_doses,
                }
            )
    df_pred = pd.DataFrame(df_list)
    uncert_cos = []
    uncert_eucl = []
    closest_cond_cos = []
    closest_cond_eucl = []
    for i in range(df_pred.shape[0]):
        uncert_cos_, uncert_eucl_, closest_cond_cos_, closest_cond_eucl_ = (
            compert_api.compute_uncertainty(
                covs=[df_pred.iloc[i]['cell_type']],
                pert=df_pred.iloc[i]['condition'],
                dose=df_pred.iloc[i]['dose_val']
            )
        )
        uncert_cos.append(uncert_cos_)
        uncert_eucl.append(uncert_eucl_)
        closest_cond_cos.append(closest_cond_cos_)
        closest_cond_eucl.append(closest_cond_eucl_)

    df_pred['uncertainty_cosine'] = uncert_cos
    df_pred['uncertainty_eucl'] = uncert_eucl
    df_pred['closest_cond_cos'] = closest_cond_cos
    df_pred['closest_cond_eucl'] = closest_cond_eucl
    doses = df_pred.dose_val.apply(lambda x: x.split('+'))
    X = np.array(
        doses
            .apply(lambda x: x[0])
            .astype(float)
    ).reshape(N, N)
    Y = np.array(
        doses
            .apply(lambda x: x[1])
            .astype(float)
    ).reshape(N, N)
    Z = np.array(
        df_pred[f'uncertainty_{metric}']
            .values
            .astype(float)
    ).reshape(N, N)

    fig, ax = plt.subplots(1, 1)
    CS = ax.contourf(X, Y, Z, cmap='coolwarm', levels=20,
                     alpha=1, vmin=vmin, vmax=vmax)

    ax.set_xlabel(pert.split('+')[0], fontweight="bold")
    ax.set_ylabel(pert.split('+')[1], fontweight="bold")
    if title:
        ax.set_title(cov)

    if not (df_ref is None):
        sns.scatterplot(
            x=pert.split('+')[0],
            y=pert.split('+')[1],
            hue='split',
            size='num_cells',
            sizes=sizes,
            alpha=1.,
            palette={'train': '#000000', 'training': '#000000', 'ood': '#e41a1c'},
            data=df_ref,
            ax=ax)
        ax.legend_.remove()

    if measured_points:
        ticks = measured_points[cov][pert]
        xticks = [float(x.split('+')[0]) for x in ticks]
        yticks = [float(x.split('+')[1]) for x in ticks]
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticks, rotation=90)
        ax.set_yticks(yticks)
    fig.colorbar(CS)
    sns.despine()
    ax.axis("equal")
    ax.axis("square")
    ax.set_xlim(xlims)
    ax.set_ylim(ylims)

    plt.tight_layout()

    if filename:
        plt.savefig(filename)

    return df_pred


def plot_uncertainty_dose(
        compet_api,
        cov,
        pert,
        N=11,
        metric='cosine',
        measured_points=None,
        cond_key='condition',
        log=False,
        min_dose=None,
        filename=None
):
    """Plotting uncertainty for a single perturbation at a dose range for a
    particular covariate.

    Params
    ------
    compert_api
        Api object for the model class.
    cov : str
        Name of covariate.
    pert : str
        Name of the perturbation.
    N : int
        Number of dose values.
    metric: str (default: 'cosine')
        Metric to evaluate uncertainty.
    measured_points : dict (default: None)
        A dicitionary of dictionaries. Per each covariate a dictionary with
        observed doses per perturbation, e.g. {'covar1': {'pert1':
        [0.1, 0.5, 1.0], 'pert2': [0.3]}
    cond_key : str (default: 'condition')
        Name of the variable to use for plotting.
    log : boolean (default: False)
        A flag if to plot on a log scale.
    min_dose : float (default: None)
        Minimum dose for the uncertainty estimate.
    filename : str (default: None)
        Full path to the file to export the plot. File extension should be included.

    Returns
        -------
        pd.DataFrame of uncertainty estimations.
    """

    df_list = []
    if log:
        if min_dose is None:
            min_dose = 1e-3
        N_val = np.round(np.logspace(np.log10(min_dose), np.log10(1), N), decimals=10)
    else:
        if min_dose is None:
            min_dose = 0
        N_val = np.round(np.linspace(min_dose, 1., N), decimals=3)

    for i in N_val:
        df_list.append(
            {
                'cell_type': cov,
                'condition': pert,
                'dose_val': i,
            }
        )
    df_pred = pd.DataFrame(df_list)
    uncert_cos = []
    uncert_eucl = []
    closest_cond_cos = []
    closest_cond_eucl = []
    for i in range(df_pred.shape[0]):
        uncert_cos_, uncert_eucl_, closest_cond_cos_, closest_cond_eucl_ = (
            compet_api.compute_uncertainty(
                covs=[df_pred.iloc[i]['cell_type']],
                pert=df_pred.iloc[i]['condition'],
                dose=df_pred.iloc[i]['dose_val']
            )
        )
        uncert_cos.append(uncert_cos_)
        uncert_eucl.append(uncert_eucl_)
        closest_cond_cos.append(closest_cond_cos_)
        closest_cond_eucl.append(closest_cond_eucl_)

    df_pred['uncertainty_cosine'] = uncert_cos
    df_pred['uncertainty_eucl'] = uncert_eucl
    df_pred['closest_cond_cos'] = closest_cond_cos
    df_pred['closest_cond_eucl'] = closest_cond_eucl

    x = df_pred.dose_val.values.astype(float)
    y = df_pred[f'uncertainty_{metric}'].values.astype(float)
    fig, ax = plt.subplots(1, 1)
    ax.plot(x, y)
    ax.set_xlabel(pert)
    ax.set_ylabel('Uncertainty')
    ax.set_title(cov)
    if log:
        ax.set_xscale('log')
    if measured_points:
        ticks = measured_points[cov][pert]
        ax.set_xticks(ticks)
        ax.set_xticklabels(ticks, rotation=90)
    else:
        plt.draw()
        ax.set_xticklabels(ax.get_xticklabels(), rotation=90)

    sns.despine()
    plt.tight_layout()

    if filename:
        plt.savefig(filename)

    return df_pred


def save_to_file(fig, file_name, file_format=None):
    if file_format is None:
        if file_name.split(".")[-1] in ['png', 'pdf']:
            file_format = file_name.split(".")[-1]
            savename = file_name
        else:
            file_format = 'pdf'
            savename = f'{file_name}.{file_format}'
    else:
        savename = file_name

    fig.savefig(savename, format=file_format)
    print(f"Saved file to: {savename}")


def plot_embedding(
        emb,
        labels=None,
        col_dict=None,
        title=None,
        show_lines=False,
        show_text=False,
        show_legend=True,
        axis_equal=True,
        circle_size=40,
        circe_transparency=1.0,
        line_transparency=0.8,
        line_width=1.0,
        fontsize=9,
        fig_width=4,
        fig_height=4,
        file_name=None,
        file_format=None,
        labels_name=None,
        width_ratios=[7, 1],
        bbox=(1.3, 0.7)
):
    sns.set_style("white")

    # create data structure suitable for embedding
    df = pd.DataFrame(emb, columns=['dim1', 'dim2'])
    if not (labels is None):
        if labels_name is None:
            labels_name = 'labels'
        df[labels_name] = labels

    fig = plt.figure(figsize=(fig_width, fig_height))
    ax = plt.gca()

    sns.despine(left=False, bottom=False, right=True)

    if (col_dict is None) and not (labels is None):
        col_dict = get_colors(labels)

    sns.scatterplot(
        x="dim1",
        y="dim2",
        hue=labels_name,
        palette=col_dict,
        alpha=circe_transparency,
        edgecolor="none",
        s=circle_size,
        data=df,
        ax=ax)

    try:
        ax.legend_.remove()
    except:
        pass

    if show_lines:
        for i in range(len(emb)):
            if col_dict is None:
                ax.plot(
                    [0, emb[i, 0]],
                    [0, emb[i, 1]],
                    alpha=line_transparency,
                    linewidth=line_width,
                    c=None
                )
            else:
                ax.plot(
                    [0, emb[i, 0]],
                    [0, emb[i, 1]],
                    alpha=line_transparency,
                    linewidth=line_width,
                    c=col_dict[labels[i]]
                )

    if show_text and not (labels is None):
        texts = []
        labels = np.array(labels)
        unique_labels = np.unique(labels)
        for label in unique_labels:
            idx_label = np.where(labels == label)[0]
            texts.append(
                ax.text(
                    np.mean(emb[idx_label, 0]),
                    np.mean(emb[idx_label, 1]),
                    label,
                    fontsize=fontsize
                )
            )

        adjust_text(
            texts,
            arrowprops=dict(arrowstyle='-', color='black', lw=0.1),
            ax=ax
        )

    if axis_equal:
        ax.axis('equal')
        ax.axis('square')

    if title:
        ax.set_title(title, fontsize=fontsize, fontweight="bold")

    ax.set_xlabel('dim1', fontsize=fontsize)
    ax.set_ylabel('dim2', fontsize=fontsize)
    ax.xaxis.set_tick_params(labelsize=fontsize)
    ax.yaxis.set_tick_params(labelsize=fontsize)

    plt.tight_layout()

    if file_name:
        save_to_file(fig, file_name, file_format)

    return plt


def get_colors(
        labels,
        palette=None,
        palette_name=None
):
    n_colors = len(labels)
    if palette is None:
        palette = get_palette(n_colors, palette_name)
    col_dict = dict(zip(labels, palette[:n_colors]))
    return col_dict


def plot_similarity(
        emb,
        labels=None,
        col_dict=None,
        fig_width=4,
        fig_height=4,
        cmap='coolwarm',
        fmt='png',
        fontsize=7,
        file_format=None,
        file_name=None
):
    # first we take construct similarity matrix
    # add another similarity
    similarity_matrix = cosine_similarity(emb)

    df = pd.DataFrame(
        similarity_matrix,
        columns=labels,
        index=labels,
    )

    if col_dict is None:
        col_dict = get_colors(labels)

    network_colors = pd.Series(df.columns, index=df.columns).map(col_dict)

    sns_plot = sns.clustermap(
        df,
        cmap=cmap,
        center=0,
        row_colors=network_colors,
        col_colors=network_colors,
        mask=False,
        metric='euclidean',
        figsize=(fig_height, fig_width),
        vmin=-1, vmax=1,
        fmt=file_format
    )

    sns_plot.ax_heatmap.xaxis.set_tick_params(labelsize=fontsize)
    sns_plot.ax_heatmap.yaxis.set_tick_params(labelsize=fontsize)
    sns_plot.ax_heatmap.axis('equal')
    sns_plot.cax.yaxis.set_tick_params(labelsize=fontsize)

    if file_name:
        save_to_file(sns_plot, file_name, file_format)


def mean_plot(
        adata,
        pred_obsm_key,
        path_to_save="./reg_mean.pdf",
        gene_list=None,
        deg_list=None,
        show=False,
        title=None,
        verbose=False,
        x_coeff=0.30,
        y_coeff=0.8,
        fontsize=11,
        R2_type="R2",
        figsize=(3.5, 3.5),
        **kwargs
):
    """
    Plots mean matching.

    # Parameters
    adata: `~anndata.AnnData`
        Contains real v
    pred_obsm_key: `str`
        Key in adata.obsm where predicted values are stored.
    condition_key: Str
        adata.obs key to look for x-axis and y-axis condition
    exp_key: Str
        Condition in adata.obs[condition_key] to be ploted
    path_to_save: basestring
        Path to save the plot.
    gene_list: list
        List of gene names to be plotted.
    deg_list: list
        List of DEGs to compute R2
    show: boolean
        if True plots the figure
    Verbose: boolean
        If true prints the value
    title: Str
        Title of the plot
    x_coeff: float
        Shifts R2 text horizontally by x_coeff
    y_coeff: float
        Shifts R2 text vertically by y_coeff
    show: bool
        if `True`: will show to the plot after saving it.
    fontsize: int
        Font size for R2 texts
    R2_type: Str
        How to compute R2 value, should be either Pearson R2 or R2 (sklearn)

    Returns:
        Calculated R2 values
    """

    r2_types = ['R2', 'Pearson R2']
    if R2_type not in r2_types:
        raise ValueError("R2 caclulation should be one of" + str(r2_types))
    if sparse.issparse(adata.X):
        adata.X = adata.X.A
    
    if deg_list is not None:
        if hasattr(deg_list, "tolist"):
            deg_list = deg_list.tolist()

        adata.layers[pred_obsm_key] = adata.obsm[pred_obsm_key].copy()
        
        deg_adata = adata[:, deg_list].copy()

        X_true_deg = np.average(deg_adata.X, axis=0)
        X_pred_deg = np.average(deg_adata.layers[pred_obsm_key], axis=0)

        if R2_type == "R2":
            r2_diff = r2_score(X_true_deg, X_pred_deg)
        if R2_type == "Pearson R2":
            m, b, pearson_r_diff, p_value_diff, std_err_diff = \
                stats.linregress(X_true_deg, X_pred_deg)
            r2_diff = pearson_r_diff ** 2
        if verbose:
            print(f'Top {len(deg_list)} DEGs var: ', r2_diff)

    x_true = np.average(adata.X, axis=0)
    x_pred = np.average(adata.obsm[pred_obsm_key], axis=0)
    
    if R2_type == "R2":
        r2 = r2_score(x_true, x_pred)
    
    elif R2_type == "Pearson R2":
        m, b, pearson_r, p_value, std_err = stats.linregress(x_true, x_pred)
        r2 = pearson_r ** 2
    
    if verbose:
        if isinstance(adata.X, np.ndarray):
            x_true = np.var(adata.X, axis=0)
        else:
            x_true = np.var(adata.X.toarray(), axis=0)
        x_pred = np.var(adata.obsm[pred_obsm_key], axis=0)

        r2_var = r2_score(x_true, x_pred)
        print('All genes var: ', r2_var)
    
    df = pd.DataFrame({f'true': x_true, f'pred': x_pred})

    plt.figure(figsize=figsize)
    ax = sns.regplot(x=f'pred', y=f'true', data=df)
    ax.tick_params(labelsize=fontsize)
    
    if "range" in kwargs:
        start, stop, step = kwargs.get("range")
        ax.set_xticks(np.arange(start, stop, step))
        ax.set_yticks(np.arange(start, stop, step))
    
    ax.set_xlabel('true', fontsize=fontsize)
    ax.set_ylabel('pred', fontsize=fontsize)
    
    if gene_list is not None:
        for i in gene_list:
            j = adata.var_names.tolist().index(i)
            x_bar = x_true[j]
            y_bar = x_pred[j]
            plt.text(x_bar, y_bar, i, fontsize=fontsize, color="black")
            plt.plot(x_bar, y_bar, 'o', color="red", markersize=5)
    
    if title is None:
        plt.title(f"", fontsize=fontsize, fontweight="bold")
    else:
        plt.title(title, fontsize=fontsize, fontweight="bold")
    
    ax.text(max(x_true) - max(x_true) * x_coeff, max(x_pred) - y_coeff * max(x_pred),
            r'$\mathrm{R^2_{\mathrm{\mathsf{all\ genes}}}}$= ' + f"{r2:.2f}",
            fontsize=fontsize)
    
    if deg_list is not None:
        ax.text(max(x_true) - max(x_true) * x_coeff, max(x_pred) - (y_coeff + 0.15) * max(x_pred),
                r'$\mathrm{R^2_{\mathrm{\mathsf{DEGs}}}}$= ' + f"{r2_diff:.2f}",
                fontsize=fontsize)
        
    plt.savefig(f"{path_to_save}", bbox_inches='tight', dpi=100)
    if show:
        plt.show()
    
    plt.close()
    
    if deg_list is not None:
        return r2, r2_diff
    else:
        return r2


def plot_r2_matrix(adata, 
                   pred_obsm_key, 
                   deg_group_key,
                   deg_uns_key: Optional[str] = None, 
                   **kwargs):
    """Plots a pairwise R2 heatmap between predicted and control conditions.

    Params
    ------
    adata : `AnnData`
        Original AnnData object
    pred_obsm_key : `str`
        Key in `adata.obsm` where predicted values are stored
    deg_group_key : `str`
        Key in `adata.obs` where DEG group information is stored
    deg_uns_key : `str`, optional (default: None)
        Key in `adata.uns` where DEG information is stored. Keys are group names 
        and values are list of DEGs for each group. If None, all genes are used.
    """
    r2s_mean = defaultdict(list)
    r2s_var = defaultdict(list)
    groups = np.unique(adata.obs[deg_group_key].values)

    for group1 in groups:
        degs = adata.uns[deg_uns_key][group1] if deg_uns_key is not None else adata.var_names
        x_pred = adata[adata.obs[deg_group_key] == group1][:, degs].obsm[pred_obsm_key]

        # calculate r2 between pairwise
        for group2 in groups:
            x_true = adata[adata.obs[deg_group_key] == group2].X
            if not isinstance(x_true, np.ndarray):
                x_true = x_true.toarray()
            r2s_mean[group2].append(r2_score(x_true.mean(axis=0), x_pred.mean(axis=0)))
            r2s_var[group2].append(r2_score(x_true.var(axis=0), x_pred.var(axis=0)))

    for r2_dict in [r2s_mean, r2s_var]:
        r2_df = pd.DataFrame.from_dict(r2_dict, orient='index')
        r2_df.columns = groups

        plt.figure(figsize=(5, 5))
        p = sns.heatmap(data=r2_df, vmin=max(r2_df.min(0).min(), -100),
                        cmap='Blues', cbar=False, square=True, xticklabels=True, yticklabels=True,
                        annot=True, fmt='.2f', annot_kws={'fontsize': 5}, **kwargs)
        plt.xticks(fontsize=6)
        plt.yticks(fontsize=6)
        plt.xlabel('y_true')
        plt.ylabel('y_pred')
        plt.show()


def plot_history(model: CPA):
    df = model.epoch_history
    n_metrics = len(df.columns) - 2
    fig, ax = plt.subplots(1, n_metrics, sharex=True, sharey=False, figsize=(24, 2.))
    for i, col in enumerate(df.columns):
        if col in ['epoch', 'mode']:
            continue

        train_df = df[df['mode'] == 'train']
        valid_df = df[df['mode'] == 'valid']

        ax[i].plot(train_df['epoch'].values, train_df[col].values, label='train')
        if len(valid_df) > 0:
            ax[i].plot(valid_df['epoch'].values, valid_df[col].values, label='valid')

        ax[i].set_title(df.columns[i + 2], fontweight="bold")

        if i == n_metrics - 1:
            handles, labels = ax[i].get_legend_handles_labels()
            fig.legend(handles, ['train', 'valid'], loc='right')



from typing import List, Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
from scvi.distributions import NegativeBinomial

from scvi.nn import FCLayers
from torch.distributions import Normal
from typing import Optional


class _REGISTRY_KEYS:
    X_KEY: str = "X"
    X_CTRL_KEY: str = None
    BATCH_KEY: str = None
    CATEGORY_KEY: str = "cpa_category"
    PERTURBATION_KEY: str = None
    PERTURBATION_DOSAGE_KEY: str = None
    PERTURBATIONS: str = "perts"
    PERTURBATIONS_DOSAGES: str = "perts_doses"
    SIZE_FACTOR_KEY: str = "size_factor"
    CAT_COV_KEYS: List[str] = []
    MAX_COMB_LENGTH: int = 2
    CONTROL_KEY: str = None
    DEG_MASK: str = None
    DEG_MASK_R2: str = None
    PADDING_IDX: int = 0


CPA_REGISTRY_KEYS = _REGISTRY_KEYS()


class VanillaEncoder(nn.Module):
    def __init__(
            self,
            n_input,
            n_output,
            n_hidden,
            n_layers,
            n_cat_list,
            use_layer_norm=True,
            use_batch_norm=False,
            output_activation: str = 'linear',
            dropout_rate: float = 0.1,
            activation_fn=nn.ReLU,
    ):
        super().__init__()
        self.n_output = n_output
        self.output_activation = output_activation

        self.network = FCLayers(
            n_in=n_input,
            n_out=n_hidden,
            n_cat_list=n_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            use_layer_norm=use_layer_norm,
            use_batch_norm=use_batch_norm,
            dropout_rate=dropout_rate,
            activation_fn=activation_fn,
        )
        self.z = nn.Linear(n_hidden, n_output)

    def forward(self, inputs, *cat_list):
        if self.output_activation == 'linear':
            z = self.z(self.network(inputs, *cat_list))
        elif self.output_activation == 'relu':
            z = F.relu(self.z(self.network(inputs, *cat_list)))
        else:
            raise ValueError(f'Unknown output activation: {self.output_activation}')
        return z


class GeneralizedSigmoid(nn.Module):
    """
    Sigmoid, log-sigmoid or linear functions for encoding dose-response for
    drug perurbations.
    """

    def __init__(self, n_drugs, non_linearity='sigmoid'):
        """Sigmoid modeling of continuous variable.
        Params
        ------
        nonlin : str (default: logsigm)
            One of logsigm, sigm.
        """
        super(GeneralizedSigmoid, self).__init__()
        self.non_linearity = non_linearity
        self.n_drugs = n_drugs

        self.beta = torch.nn.Parameter(
            torch.ones(1, n_drugs),
            requires_grad=True
        )
        self.bias = torch.nn.Parameter(
            torch.zeros(1, n_drugs),
            requires_grad=True
        )

        self.vmap = None

    def forward(self, x, y):
        """
            Parameters
            ----------
            x: (batch_size, max_comb_len)
            y: (batch_size, max_comb_len)
        """
        y = y.long()
        if self.non_linearity == 'logsigm':
            bias = self.bias[0][y]
            beta = self.beta[0][y]
            c0 = bias.sigmoid()
            return (torch.log1p(x) * beta + bias).sigmoid() - c0
        elif self.non_linearity == 'sigm':
            bias = self.bias[0][y]
            beta = self.beta[0][y]
            c0 = bias.sigmoid()
            return (x * beta + bias).sigmoid() - c0
        else:
            return x

    def one_drug(self, x, i):
        if self.non_linearity == 'logsigm':
            c0 = self.bias[0][i].sigmoid()
            return (torch.log1p(x) * self.beta[0][i] + self.bias[0][i]).sigmoid() - c0
        elif self.non_linearity == 'sigm':
            c0 = self.bias[0][i].sigmoid()
            return (x * self.beta[0][i] + self.bias[0][i]).sigmoid() - c0
        else:
            return x


class PerturbationNetwork(nn.Module):
    def __init__(self,
                 n_perts,
                 n_latent,
                 doser_type='logsigm',
                 n_hidden=None,
                 n_layers=None,
                 dropout_rate: float = 0.0,
                 drug_embeddings=None,):
        super().__init__()
        self.n_latent = n_latent
        
        if drug_embeddings is not None:
            self.pert_embedding = drug_embeddings
            self.pert_transformation = nn.Linear(drug_embeddings.embedding_dim, n_latent)
            self.use_rdkit = True
        else:
            self.use_rdkit = False
            self.pert_embedding = nn.Embedding(n_perts, n_latent, padding_idx=CPA_REGISTRY_KEYS.PADDING_IDX)
            
        self.doser_type = doser_type
        if self.doser_type == 'mlp':
            self.dosers = nn.ModuleList()
            for _ in range(n_perts):
                self.dosers.append(
                    FCLayers(
                        n_in=1,
                        n_out=1,
                        n_hidden=n_hidden,
                        n_layers=n_layers,
                        use_batch_norm=False,
                        use_layer_norm=True,
                        dropout_rate=dropout_rate
                    )
                )
        else:
            self.dosers = GeneralizedSigmoid(n_perts, non_linearity=self.doser_type)

    def forward(self, perts, dosages):
        """
            perts: (batch_size, max_comb_len)
            dosages: (batch_size, max_comb_len)
        """
        bs, max_comb_len = perts.shape
        perts = perts.long()
        scaled_dosages = self.dosers(dosages, perts)  # (batch_size, max_comb_len)

        drug_embeddings = self.pert_embedding(perts)  # (batch_size, max_comb_len, n_drug_emb_dim)

        if self.use_rdkit:
            drug_embeddings = self.pert_transformation(drug_embeddings.view(bs * max_comb_len, -1)).view(bs, max_comb_len, -1)

        z_drugs = torch.einsum('bm,bme->bme', [scaled_dosages, drug_embeddings])  # (batch_size, n_latent)

        z_drugs = torch.einsum('bmn,bm->bmn', z_drugs, (perts != CPA_REGISTRY_KEYS.PADDING_IDX).int()).sum(dim=1)  # mask single perts

        return z_drugs # (batch_size, n_latent)

class FocalLoss(nn.Module):
    """ Inspired by https://github.com/AdeelH/pytorch-multi-class-focal-loss/blob/master/focal_loss.py

    Focal Loss, as described in https://arxiv.org/abs/1708.02002.
    It is essentially an enhancement to cross entropy loss and is
    useful for classification tasks when there is a large class imbalance.
    x is expected to contain raw, unnormalized scores for each class.
    y is expected to contain class labels.
    Shape:
        - x: (batch_size, C) or (batch_size, C, d1, d2, ..., dK), K > 0.
        - y: (batch_size,) or (batch_size, d1, d2, ..., dK), K > 0.
    """

    def __init__(self,
                 alpha: Optional[torch.Tensor] = None,
                 gamma: float = 2.,
                 reduction: str = 'mean',
                 ):
        """
        Args:
            alpha (Tensor, optional): Weights for each class. Defaults to None.
            gamma (float, optional): A constant, as described in the paper.
                Defaults to 0.
            reduction (str, optional): 'mean', 'sum' or 'none'.
                Defaults to 'mean'.
        """
        if reduction not in ('mean', 'sum', 'none'):
            raise ValueError(
                'Reduction must be one of: "mean", "sum", "none".')

        super().__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction

        self.nll_loss = nn.NLLLoss(
            weight=alpha, reduction='none')

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        if len(y_true) == 0:
            return torch.tensor(0.)

        # compute weighted cross entropy term: -alpha * log(pt)
        # (alpha is already part of self.nll_loss)
        log_p = F.log_softmax(y_pred, dim=-1)
        ce = self.nll_loss(log_p, y_true)

        # get true class column from each row
        all_rows = torch.arange(len(y_pred))
        log_pt = log_p[all_rows, y_true]

        # compute focal term: (1 - pt)^gamma
        pt = log_pt.exp()
        focal_term = (1 - pt) ** self.gamma

        # the full loss: -alpha * ((1 - pt)^gamma) * log(pt)
        loss = focal_term * ce

        if self.reduction == 'mean':
            loss = loss.mean()
        elif self.reduction == 'sum':
            loss = loss.sum()

        return loss


import json
import logging
import os
from tkinter import N
from typing import Optional, Sequence, Union, List, Dict

from rdkit import Chem
from rdkit.Chem import AllChem

import torch.nn as nn

import numpy as np
import pandas as pd
import torch
from pytorch_lightning.callbacks import EarlyStopping
from scvi.data import AnnDataManager
from scvi.dataloaders import DataSplitter
from torch.nn import functional as F
from scvi.data.fields import (
    LayerField,
    CategoricalObsField,
    NumericalObsField,
    ObsmField,
)

from anndata import AnnData
from scvi.model.base import BaseModelClass
from scvi.train import TrainRunner
from scvi.train._callbacks import SaveBestState
from scvi.utils import setup_anndata_dsp
from tqdm import tqdm

from ._module import CPAModule
from ._utils import CPA_REGISTRY_KEYS
from ._task import CPATrainingPlan
from ._data import AnnDataSplitter

logger = logging.getLogger(__name__)
logger.propagate = False


class CPA(BaseModelClass):
    """CPA model

    Parameters
    ----------
    adata : Anndata
        Registered Annotation Dataset

    n_latent: int
        Number of latent dimensions used for drug and Autoencoder

    recon_loss: str
        Either `gauss` or `NB`. Autoencoder loss function.

    doser_type: str
        Type of doser network. Either `sigm`, `logsigm` or `mlp`.

    split_key : str, optional
        Key used to split the data between train test and validation.
        This must correspond to a observation key for the adata, composed of values
        'train', 'test', and 'ood'. By default None.

    **hyper_params:
        CPA's hyper-parameters.


    Examples
    --------
    >>> import cpa
    >>> import scanpy as sc
    >>> adata = sc.read('dataset.h5ad')
    >>> adata = cpa.CPA.setup_anndata(adata,
                                      perturbation_keys={'perturbation': 'condition',  'dosage': 'dose_val'},
                                      categorical_covariate_keys=['cell_type'],
                                      control_key='control'
                                      )
    >>> hyperparams = {'autoencoder_depth': 3, 'autoencoder_width': 256}
    >>> model = cpa.CPA(adata,
                        n_latent=256,
                        loss_ae='gauss',
                        doser_type='logsigm',
                        split_key='split',
                        )
    """

    covars_encoder: dict = None
    pert_encoder: dict = None
    pert_smiles_map: dict = None

    def __init__(
        self,
        adata: AnnData,
        split_key: str = None,
        train_split: Union[str, List[str]] = "train",
        valid_split: Union[str, List[str]] = "test",
        test_split: Union[str, List[str]] = "ood",
        use_rdkit_embeddings: bool = False,
        **hyper_params,
    ):
        super().__init__(adata)

        self.split_key = split_key

        self.drugs = list(self.pert_encoder.keys())
        self.covars = {
            covar: list(self.covars_encoder[covar].keys())
            for covar in self.covars_encoder.keys()
        }

        if use_rdkit_embeddings and self.pert_smiles_map is not None:
            # get morgan fingerprint vectors for drug embeddings
            drug_embeddings = self.__get_rdkit_embeddings()
            hyper_params['drug_embeddings'] = drug_embeddings

        self.module = CPAModule(
            n_genes=adata.n_vars,
            n_perts=len(self.pert_encoder),
            covars_encoder=self.covars_encoder,
            **hyper_params,
        ).float()

        train_indices, valid_indices, test_indices = None, None, None
        if split_key is not None:
            train_split = (
                train_split if isinstance(train_split, list) else [train_split]
            )
            valid_split = (
                valid_split if isinstance(valid_split, list) else [valid_split]
            )
            test_split = test_split if isinstance(test_split, list) else [test_split]

            train_indices = np.where(adata.obs.loc[:, split_key].isin(train_split))[0]
            valid_indices = np.where(adata.obs.loc[:, split_key].isin(valid_split))[0]
            test_indices = np.where(adata.obs.loc[:, split_key].isin(test_split))[0]

        self.train_indices = train_indices
        self.valid_indices = valid_indices
        self.test_indices = test_indices

        self._model_summary_string = f"Compositional Perturbation Autoencoder"

        self.init_params_ = self._get_init_params(locals())

        self.epoch_history = None

    def __get_rdkit_embeddings(
        self, 
    ):
        assert self.pert_smiles_map not in [None, []]
        query_drug_names = list(self.pert_encoder.keys())
        query_drug_names.remove('<PAD>')

        smiles_list = [self.pert_smiles_map[drug] for drug in list(query_drug_names)]

        drug_fps = []
        for smiles in smiles_list:
            mol = Chem.MolFromSmiles(smiles)
            fps = AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=2048)
            drug_fps.append(np.array(fps))
        
        drug_fps = np.vstack(drug_fps)

        print(drug_fps.shape)

        embeddings = AnnData(X=drug_fps)
        embeddings.obs.index = smiles_list
        embeddings = embeddings[list(smiles_list), :]

        drug_embeddings = nn.Embedding(
            len(self.pert_encoder),
            embeddings.shape[1],
            padding_idx=CPA_REGISTRY_KEYS.PADDING_IDX,
        )
        pad_X = np.zeros(shape=(1, embeddings.n_vars))
        X = np.concatenate((pad_X, embeddings.X), 0)
        drug_embeddings.weight.data.copy_(torch.tensor(X))
        drug_embeddings.weight.requires_grad = False

        return drug_embeddings

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        perturbation_key: str,
        control_group: str,
        dosage_key: Optional[str] = None,
        batch_key: Optional[str] = None,
        layer: Optional[str] = None,
        smiles_key: Optional[str] = None,
        is_count_data: Optional[bool] = True,
        categorical_covariate_keys: Optional[List[str]] = [],
        deg_uns_key: Optional[str] = None,
        deg_uns_cat_key: Optional[str] = None,
        max_comb_len: int = 2,
        **kwargs,
    ):
        """
        Annotation Data setup function

        Parameters
        ----------
        adata: anndata.AnnData
            AnnData object
        perturbation_key: str
            Key in `adata.obs` containing perturbations
        control_group: str
            Control group name
        dosage_key: str, optional
            Key in `adata.obs` containing perturbation dosages, by default None. If None, all dosages are set to 1.0
        batch_key: str, optional
            Key in `adata.obs` containing batch information, by default None
        layer: str, optional
            Key in `adata.layers` containing gene expression data, by default None. If None, `adata.X` is used
        is_count_data: bool, optional
            Whether the data is count data, by default False
        categorical_covariate_keys: List[str], optional
            List of keys in `adata.obs` containing categorical covariates, by default None
        deg_uns_key: str, optional
            Key in `adata.uns` containing differentially expressed genes for each combination of covariates and perturbations, by default None
        deg_uns_cat_key: str, optional
            Key in `adata.obs` containing covariate combinations for each cell, by default None
        max_comb_len: int, optional
            Maximum number of perturbations in a combination, by default 2
        """
        CPA_REGISTRY_KEYS.PERTURBATION_KEY = perturbation_key
        CPA_REGISTRY_KEYS.PERTURBATION_DOSAGE_KEY = dosage_key
        CPA_REGISTRY_KEYS.CAT_COV_KEYS = categorical_covariate_keys
        CPA_REGISTRY_KEYS.MAX_COMB_LENGTH = max_comb_len
        CPA_REGISTRY_KEYS.BATCH_KEY = batch_key

        if dosage_key is None:
            print(f'Warning: dosage_key is not set. Setting it to "1.0" for all cells')

            dosage_key = 'CPA_dose_val'

            adata.obs[dosage_key] = adata.obs[perturbation_key].apply(lambda x: '+'.join(['1.0' for _ in x.split('+')])).values

        CPA_REGISTRY_KEYS.PERTURBATION_DOSAGE_KEY = dosage_key

        perturbations = adata.obs[perturbation_key].astype(str).values
        dosages = adata.obs[dosage_key].astype(str).values

        category_key = f"{cls.__name__}_cat"
        keys = categorical_covariate_keys + [perturbation_key]

        if batch_key is not None:
            keys = [batch_key] + keys

        adata.obs[category_key] = adata.obs[keys].apply(lambda x: "_".join(x), axis=1)
        CPA_REGISTRY_KEYS.CATEGORY_KEY = category_key

        if cls.pert_encoder is None:
            # get unique drugs
            perts_names_unique = set()
            for d in np.unique(perturbations):
                [perts_names_unique.add(i) for i in d.split("+") if i != control_group]
            perts_names_unique = ["<PAD>", control_group] + sorted(
                list(perts_names_unique)
            )
            CPA_REGISTRY_KEYS.PADDING_IDX = 0

            pert_encoder = {pert: i for i, pert in enumerate(perts_names_unique)}

        else:
            pert_encoder = cls.pert_encoder
            perts_names_unique = list(pert_encoder.keys())

        if smiles_key is not None:
            if cls.pert_smiles_map is None:
                pert_smiles_map = {}
                for pert in perts_names_unique:
                    if pert != "<PAD>":
                        try:
                            pert_smiles_map[pert] = adata.obs.loc[
                                adata.obs[perturbation_key] == pert, smiles_key
                            ].values[0]
                        except:
                            pert_name = adata.obs.loc[
                                adata.obs[perturbation_key].str.contains(pert), perturbation_key
                            ].values[0]

                            smiles = adata.obs.loc[
                                adata.obs[perturbation_key].str.contains(pert), smiles_key
                            ].values[0]

                            pert_smiles_map[pert] = smiles.split('..')[pert_name.split('+').index(pert)]
                cls.pert_smiles_map = pert_smiles_map
            else:
                pert_smiles_map = cls.pert_smiles_map

        pert_map = {}
        for condition in tqdm(perturbations):
            perts_list = np.where(np.isin(perts_names_unique, condition.split("+")))[0]
            pert_map[condition] = list(perts_list) + [
                CPA_REGISTRY_KEYS.PADDING_IDX
                for _ in range(max_comb_len - len(perts_list))
            ]

        dose_map = {}
        for dosage_str in tqdm(dosages):
            dosages_list = [float(i) for i in dosage_str.split("+")]
            dose_map[dosage_str] = list(dosages_list) + [
                0.0 for _ in range(max_comb_len - len(dosages_list))
            ]

        data_perts = np.vstack(
            np.vectorize(lambda x: pert_map[x], otypes=[np.ndarray])(perturbations)
        ).astype(int)
        adata.obsm[CPA_REGISTRY_KEYS.PERTURBATIONS] = data_perts

        data_perts_dosages = np.vstack(
            np.vectorize(lambda x: dose_map[x], otypes=[np.ndarray])(dosages)
        ).astype(float)
        adata.obsm[CPA_REGISTRY_KEYS.PERTURBATIONS_DOSAGES] = data_perts_dosages

        # setup control column
        control_key = f"{cls.__name__}_{control_group}"
        CPA_REGISTRY_KEYS.CONTROL_KEY = control_key
        adata.obs[control_key] = (adata.obs[perturbation_key] == control_group).astype(
            int
        )

        setup_method_args = cls._get_setup_method_args(**locals())
        anndata_fields = [
            LayerField(
                registry_key=CPA_REGISTRY_KEYS.X_KEY,
                layer=layer,
                is_count_data=is_count_data,
            ),
            ObsmField(
                CPA_REGISTRY_KEYS.PERTURBATIONS,
                CPA_REGISTRY_KEYS.PERTURBATIONS,
                is_count_data=True,
                correct_data_format=True,
            ),
            ObsmField(
                CPA_REGISTRY_KEYS.PERTURBATIONS_DOSAGES,
                CPA_REGISTRY_KEYS.PERTURBATIONS_DOSAGES,
                is_count_data=False,
                correct_data_format=True,
            ),
            CategoricalObsField(
                registry_key=CPA_REGISTRY_KEYS.PERTURBATION_KEY,
                attr_key=perturbation_key,
            ),
        ] + [
            CategoricalObsField(registry_key=covar, attr_key=covar)
            for covar in categorical_covariate_keys
        ]

        anndata_fields.append(
            NumericalObsField(registry_key=control_key, attr_key=control_key)
        )
        anndata_fields.append(
            CategoricalObsField(registry_key=category_key, attr_key=category_key)
        )

        if batch_key is not None:
            anndata_fields.append(
                CategoricalObsField(registry_key=batch_key, attr_key=batch_key)
            )

        if deg_uns_key:
            n_deg_r2 = kwargs.pop("n_deg_r2", 10)

            cov_cond_unique = np.unique(adata.obs[deg_uns_cat_key].astype(str).values)

            cov_cond_map = {}
            cov_cond_map_r2 = {}
            for cov_cond in tqdm(cov_cond_unique):
                if cov_cond in adata.uns[deg_uns_key].keys():
                    mask_hvg = adata.var_names.isin(
                        adata.uns[deg_uns_key][cov_cond]
                    ).astype(np.int)
                    mask_hvg_r2 = adata.var_names.isin(
                        adata.uns[deg_uns_key][cov_cond][:n_deg_r2]
                    ).astype(np.int)
                    cov_cond_map[cov_cond] = list(mask_hvg)
                    cov_cond_map_r2[cov_cond] = list(mask_hvg_r2)
                else:
                    no_mask = list(np.ones(shape=(adata.n_vars,)))
                    cov_cond_map[cov_cond] = no_mask
                    cov_cond_map_r2[cov_cond] = no_mask

            mask = np.vstack(
                np.vectorize(lambda x: cov_cond_map[x], otypes=[np.ndarray])(
                    adata.obs[deg_uns_cat_key].astype(str).values
                )
            )
            mask_r2 = np.vstack(
                np.vectorize(lambda x: cov_cond_map[x], otypes=[np.ndarray])(
                    adata.obs[deg_uns_cat_key].astype(str).values
                )
            )

            CPA_REGISTRY_KEYS.DEG_MASK = "deg_mask"
            CPA_REGISTRY_KEYS.DEG_MASK_R2 = "deg_mask_r2"
            adata.obsm[CPA_REGISTRY_KEYS.DEG_MASK] = np.array(mask)
            adata.obsm[CPA_REGISTRY_KEYS.DEG_MASK_R2] = np.array(mask_r2)

            anndata_fields.append(
                ObsmField(
                    CPA_REGISTRY_KEYS.DEG_MASK,
                    CPA_REGISTRY_KEYS.DEG_MASK,
                    is_count_data=True,
                    correct_data_format=True,
                )
            )
            anndata_fields.append(
                ObsmField(
                    CPA_REGISTRY_KEYS.DEG_MASK_R2,
                    CPA_REGISTRY_KEYS.DEG_MASK_R2,
                    is_count_data=True,
                    correct_data_format=True,
                )
            )

        adata_manager = AnnDataManager(
            fields=anndata_fields, setup_method_args=setup_method_args
        )
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

        keys = categorical_covariate_keys
        if batch_key is not None:
            keys.append(batch_key)

        covars_encoder = {}
        for covar in keys:
            covars_encoder[covar] = {
                c: i
                for i, c in enumerate(
                    adata_manager.registry["field_registries"][covar]["state_registry"][
                        "categorical_mapping"
                    ]
                )
            }

        if cls.covars_encoder is None:
            cls.covars_encoder = covars_encoder

        if cls.pert_encoder is None:
            cls.pert_encoder = pert_encoder

    def train(
        self,
        max_epochs: Optional[int] = None,
        use_gpu: Optional[Union[str, int, bool]] = None,
        train_size: float = 0.9,
        validation_size: Optional[float] = None,
        batch_size: int = 128,
        plan_kwargs: Optional[dict] = None,
        save_path: Optional[str] = None,
        check_val_every_n_epoch: int = 10,
        early_stopping_patience: int = 10,
        **trainer_kwargs,
    ):
        """
        Trains CPA on the given dataset

        Parameters
        ----------
        max_epochs: int
            Maximum number of epochs for training
        use_gpu: bool
            Whether to use GPU if available
        train_size: float
            Fraction of training data in the case of randomly splitting dataset to train/valdiation
                if `split_key` is not set in model's constructor
        validation_size: float
            Fraction of validation data in the case of randomly splitting dataset to train/valdiation
                if `split_key` is not set in model's constructor
        batch_size: int
            Size of mini-batches for training
        early_stopping: bool
            If `True`, EarlyStopping will be used during training on validation dataset
        plan_kwargs: dict
            `CPATrainingPlan` parameters
        save_path: str
            Path to save the model after the end of training
        check_val_every_n_epoch: int
            How often to check validation metrics
        early_stopping_patience: int
            Patience for early stopping
        **trainer_kwargs:
            Additional parameters for `cpa.CPATrainingPlan`
        """
        if max_epochs is None:
            n_cells = self.adata.n_obs
            max_epochs = np.min([round((20000 / n_cells) * 400), 400])
        plan_kwargs = plan_kwargs if isinstance(plan_kwargs, dict) else dict()

        manual_splitting = (
            (self.valid_indices is not None)
            and (self.train_indices is not None)
            and (self.test_indices is not None)
        )
        if manual_splitting:
            data_splitter = AnnDataSplitter(
                self.adata_manager,
                train_indices=self.train_indices,
                valid_indices=self.valid_indices,
                test_indices=self.test_indices,
                batch_size=batch_size,
                use_gpu=use_gpu,
            )
        else:
            data_splitter = DataSplitter(
                self.adata_manager,
                train_size=train_size,
                validation_size=validation_size,
                batch_size=batch_size,
                use_gpu=use_gpu,
            )

        perturbation_key = CPA_REGISTRY_KEYS.PERTURBATION_KEY
        pert_adv_encoder = {
            c: i
            for i, c in enumerate(
                self.adata_manager.registry["field_registries"][perturbation_key][
                    "state_registry"
                ]["categorical_mapping"]
            )
        }

        drug_weights = []
        n_adv_perts = len(self.adata.obs[perturbation_key].unique())
        for condition in tqdm(list(pert_adv_encoder.keys())):
            n_positive = len(self.adata[self.adata.obs[perturbation_key] == condition])
            drug_weights.append((self.adata.n_obs / n_positive) - 1.0)

        self.training_plan = CPATrainingPlan(
            self.module,
            self.covars_encoder,
            n_adv_perts=n_adv_perts,
            **plan_kwargs,
            drug_weights=drug_weights,
        )
        trainer_kwargs["early_stopping"] = False
        trainer_kwargs["check_val_every_n_epoch"] = check_val_every_n_epoch

        es_callback = EarlyStopping(
            monitor="cpa_metric",
            patience=early_stopping_patience,
            check_on_train_epoch_end=False,
            verbose=False,
            mode="max",
        )

        if "callbacks" in trainer_kwargs.keys() and isinstance(
            trainer_kwargs.get("callbacks"), list
        ):
            trainer_kwargs["callbacks"] += [es_callback]
        else:
            trainer_kwargs["callbacks"] = [es_callback]

        if save_path is None:
            save_path = "./"

        checkpoint = SaveBestState(
            monitor="cpa_metric", mode="max", period=1, verbose=True
        )
        trainer_kwargs["callbacks"].append(checkpoint)

        self.runner = TrainRunner(
            self,
            training_plan=self.training_plan,
            data_splitter=data_splitter,
            max_epochs=max_epochs,
            use_gpu=use_gpu,
            early_stopping_monitor="cpa_metric",
            early_stopping_mode="max",
            **trainer_kwargs,
        )
        self.runner()

        self.epoch_history = pd.DataFrame().from_dict(self.training_plan.epoch_history)
        if save_path is not False:
            self.save(save_path, overwrite=True)

    @torch.no_grad()
    def get_latent_representation(
        self,
        adata: Optional[AnnData] = None,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = 32,
    ):
        """Returns All latent representations for the given dataset

        Parameters
        ----------
        adata : Optional[AnnData], optional
            [description], by default None
        indices : Optional[Sequence[int]], optional
            Optional indices, by default None
        batch_size : Optional[int], optional
            Batch size to use, by default None

        Returns
        -------
        latent_outputs : Dict[str, anndata.AnnData]
            Dictionary of latent representations containing:
                - 'latent_corrected': batch-corrected (if batch_key is set) latent representation
                - 'latent_basal': basal latent representation
                - 'latent_after': final latent representation which can be used for decoding.
        """

        if self.is_trained_ is False:
            raise RuntimeError("Please train the model first.")

        adata = self._validate_anndata(adata)
        if indices is None:
            indices = np.arange(adata.n_obs)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size, shuffle=False
        )

        latent_basal = []
        latent = []
        latent_corrected = []
        for tensors in tqdm(scdl):
            tensors, _ = self.module.mixup_data(tensors, alpha=0.0)
            inference_inputs = self.module._get_inference_input(tensors)
            outputs = self.module.inference(**inference_inputs)
            latent_basal += [outputs["z_basal"].cpu().numpy()]
            latent += [outputs["z"].cpu().numpy()]
            latent_corrected += [outputs["z_corrected"].cpu().numpy()]

        latent_basal_adata = AnnData(
            X=np.concatenate(latent_basal, axis=0), obs=adata.obs.copy()
        )
        latent_basal_adata.obs_names = adata.obs_names

        latent_corrected_adata = AnnData(
            X=np.concatenate(latent_corrected, axis=0), obs=adata.obs.copy()
        )
        latent_corrected_adata.obs_names = adata.obs_names

        latent_adata = AnnData(X=np.concatenate(latent, axis=0), obs=adata.obs.copy())
        latent_adata.obs_names = adata.obs_names

        latent_outputs = {
            "latent_corrected": latent_corrected_adata,
            "latent_basal": latent_basal_adata,
            "latent_after": latent_adata,
        }

        return latent_outputs

    @torch.no_grad()
    def predict(
        self,
        adata: Optional[AnnData] = None,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = 32,
        n_samples: int = 20,
        return_mean: bool = True,
    ):
        """Counterfactual-friendly gene expression prediction

        To produce counterfactuals, save `adata.X` to `adata.obsm['X_true']` 
        and set it to control cells gene expression.

        For the case of reconstruction, you can pass original adata 
        without any further modifications.

        Returns
        -------
        None (predictions are saved to `adata.obsm[f'CPA_pred']`)
        """
        assert self.module.recon_loss in ["gauss", "nb", "zinb"]
        self.module.eval()

        adata = self._validate_anndata(adata)
        if indices is None:
            indices = np.arange(adata.n_obs)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size, shuffle=False
        )
        xs = []
        for tensors in tqdm(scdl):
            x_pred = (
                self.module.get_expression(tensors, n_samples=n_samples)['px']
                .detach()
                .cpu()
                .numpy()
            )
            xs.append(x_pred)

        if n_samples > 1 and self.module.variational:
            # The -2 axis correspond to cells.
            x_pred = np.concatenate(xs, axis=1)
        else:
            x_pred = np.concatenate(xs, axis=0)

        if self.module.variational and n_samples > 1 and return_mean:
            x_pred = x_pred.mean(0)

        adata.obsm[f"{self.__class__.__name__}_pred"] = x_pred

    def custom_predict(
        self,
        covars_to_add: Optional[Sequence[str]] = None,
        basal=False,
        add_batch: bool = True,
        add_pert: bool = True,
        adata: Optional[AnnData] = None,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = 32,
        n_samples: int = 20,
        return_mean: bool = True,
    ) -> AnnData:
        """
        Predicts the output of the model on the given input data.

        Args:
            covars_to_add (Optional[Sequence[str]]): List of covariates to add to the basal latent representation.
            basal (bool): Whether to use just the basal latent representation. If True, `add_batch` and `add_pert` are ignored.
            add_batch (bool): Whether to add the batch covariate to the latent representation.
            add_pert (bool): Whether to add the perturbation covariate to the latent representation.
            adata (Optional[AnnData]): The input data to predict on.
            indices (Optional[Sequence[int]]): The indices of the cells to predict on.
            batch_size (Optional[int]): The batch size to use for prediction.
            n_samples (int): The number of samples to use for stochastic prediction.
            return_mean (bool): Whether to return the mean of the samples or all the samples.

        Returns:
            latent_outputs (AnnData): A dictionary of AnnData objects containing the predicted gene expression for the specified 
            covariates, and latent representations for different covariate combinations.
        """
        if covars_to_add is None:
            covars_to_add = []
        for covar in covars_to_add:
            assert covar in self.module.covars_encoder.keys(
            ), f"covariate {covar} not found in learned covariates"

        if basal:
            latent_key = "z_basal"
        else:
            if add_batch and add_pert:
                latent_key = "z"
            elif add_batch:
                latent_key = "z_no_pert"
            elif add_pert:
                latent_key = "z_corrected"
            else:
                latent_key = "z_no_pert_corrected"

        assert self.module.recon_loss in ["gauss", "nb", "zinb"]
        self.module.eval()

        adata = self._validate_anndata(adata)
        if indices is None:
            indices = np.arange(adata.n_obs)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size, shuffle=False
        )
        xs = []
        zs = []
        z_correcteds = []
        z_no_perts = []
        z_no_pert_correcteds = []
        z_basals = []
        for tensors in tqdm(scdl):
            predictions = self.module.get_expression(
                tensors, n_samples=n_samples, covars_to_add=covars_to_add, latent=latent_key)

            px = predictions['px']
            z = predictions['z']
            z_corrected = predictions['z_corrected']
            z_no_pert = predictions['z_no_pert']
            z_no_pert_corrected = predictions['z_no_pert_corrected']
            z_basal = predictions['z_basal']

            x_pred = (
                px.detach().cpu().numpy()
            )
            xs.append(x_pred)

            z = (
                z.detach().cpu().numpy()
            )
            zs.append(z)

            z_corrected = (
                z_corrected.detach().cpu().numpy()
            )
            z_correcteds.append(z_corrected)

            z_no_pert = (
                z_no_pert.detach().cpu().numpy()
            )
            z_no_perts.append(z_no_pert)

            z_no_pert_corrected = (
                z_no_pert_corrected.detach().cpu().numpy()
            )
            z_no_pert_correcteds.append(z_no_pert_corrected)

            z_basal = (
                z_basal.detach().cpu().numpy()
            )
            z_basals.append(z_basal)

        if n_samples > 1 and self.module.variational:
            # The -2 axis correspond to cells.
            x_pred = np.concatenate(xs, axis=1)
            z = np.concatenate(zs, axis=1)
            z_corrected = np.concatenate(z_correcteds, axis=1)
            z_no_pert = np.concatenate(z_no_perts, axis=1)
            z_no_pert_corrected = np.concatenate(z_no_pert_correcteds, axis=1)
            z_basal = np.concatenate(z_basals, axis=1)
        else:
            x_pred = np.concatenate(xs, axis=0)
            z = np.concatenate(zs, axis=0)
            z_corrected = np.concatenate(z_correcteds, axis=0)
            z_no_pert = np.concatenate(z_no_perts, axis=0)
            z_no_pert_corrected = np.concatenate(z_no_pert_correcteds, axis=0)
            z_basal = np.concatenate(z_basals, axis=0)

        if self.module.variational and n_samples > 1 and return_mean:
            x_pred = x_pred.mean(0)
            z = z.mean(0)
            z_corrected = z_corrected.mean(0)
            z_no_pert = z_no_pert.mean(0)
            z_no_pert_corrected = z_no_pert_correcteds.mean(0)
            z_basal = z_basal.mean(0)

        latent_x_pred = AnnData(
            X=x_pred, obs=adata.obs.copy()
        )
        latent_x_pred.obs_names = adata.obs_names

        latent_z = AnnData(
            X=z, obs=adata.obs.copy()
        )
        latent_z.obs_names = adata.obs_names

        latent_z_corrected = AnnData(
            X=z_corrected, obs=adata.obs.copy()
        )
        latent_z_corrected.obs_names = adata.obs_names

        latent_z_no_pert = AnnData(
            X=z_no_pert, obs=adata.obs.copy()
        )
        latent_z_no_pert.obs_names = adata.obs_names

        latent_z_no_pert_corrected = AnnData(
            X=z_no_pert_corrected, obs=adata.obs.copy()
        )
        latent_z_no_pert_corrected.obs_names = adata.obs_names

        latent_z_basal = AnnData(
            X=z_basal, obs=adata.obs.copy()
        )
        latent_z_basal.obs_names = adata.obs_names

        latent_outputs = {
            "latent_x_pred": latent_x_pred,
            "latent_z": latent_z,
            "latent_z_corrected": latent_z_corrected,
            "latent_z_no_pert": latent_z_no_pert,
            "latent_z_no_pert_corrected": latent_z_no_pert_corrected,
            "latent_z_basal": latent_z_basal,
        }

        return latent_outputs

    @torch.no_grad()
    def get_pert_embeddings(self, dosage=1.0, pert: Optional[str] = None):
        """Computes all/specific perturbation (e.g. drug) embeddings

        Parameters
        ----------
        dosage : float
            Dosage of interest, by default 1.0
        pert: str, optional
            Perturbation name if single perturbation embedding is desired

        Returns
        -------
        AnnData with perturbation embeddings in `.X` and perturbation names saved in `.obs['pert_name']`.

        """
        self.module.eval()
        if isinstance(dosage, float):
            if pert is None:
                n_drugs = len(self.pert_encoder)
                treatments = [torch.arange(n_drugs, device=self.device).long().unsqueeze(1)]
                for _ in range(CPA_REGISTRY_KEYS.MAX_COMB_LENGTH - 1):
                    treatments += [torch.zeros(n_drugs, device=self.device).long().unsqueeze(1) + CPA_REGISTRY_KEYS.PADDING_IDX]
                
                treatments = torch.cat(treatments, dim=1) # (n_drugs, max_comb_len)
                treatments_dosages = [torch.tensor([dosage for _ in range(n_drugs)], device=self.device).float().unsqueeze(1)] # (n_drugs, 1)
                for _ in range(CPA_REGISTRY_KEYS.MAX_COMB_LENGTH - 1):
                    treatments_dosages += [torch.zeros(n_drugs, device=self.device).float().unsqueeze(1) + CPA_REGISTRY_KEYS.PADDING_IDX]
                treatments_dosages = torch.cat(treatments_dosages, dim=1) # (n_drugs, max_comb_len)
            else:
                treatments = [self.pert_encoder[pert]] + [CPA_REGISTRY_KEYS.PADDING_IDX for _ in range(CPA_REGISTRY_KEYS.MAX_COMB_LENGTH - 1)]
                treatments = torch.LongTensor(treatments).to(self.device).unsqueeze(0)

                treatments_dosages = [dosage] + [CPA_REGISTRY_KEYS.PADDING_IDX for _ in range(CPA_REGISTRY_KEYS.MAX_COMB_LENGTH - 1)]
                treatments_dosages = torch.FloatTensor(treatments_dosages).to(self.device).unsqueeze(0)
        else:
            raise NotImplementedError

        embeds = self.module.pert_network(treatments, treatments_dosages).detach().cpu().numpy() # (1 or n_drugs, n_latent)
        pert_latent_adata = AnnData(X=embeds)
        pert_latent_adata.obs['pert_name'] = [pert] if pert is not None else self.pert_encoder.keys()

        return pert_latent_adata

    @torch.no_grad()
    def get_covar_embeddings(self, covariate: str, covariate_value: str = None):
        """Computes Covariate embeddings (e.g. cell_type, tissue, etc.)

        Parameters
        ----------
        covariate : str
            covariate to be computed
        covariate_value: str, Optional
            Covariate specific value for embedding computation

        Returns
        -------
        AnnData object with covariate embeddings in `.X` and covariate values in `.obs[covariate]`

        """
        # assert and print the error 
        assert covariate in self.covars_encoder.keys(), f"covariate {covariate} not found in learned covariates"
        self.module.eval()

        if covariate_value is None:
            covar_ids = torch.arange(
                len(self.covars_encoder[covariate]), device=self.device
            ).long().unsqueeze(1)
        else:
            covar_ids = torch.LongTensor(
                [self.covars_encoder[covariate][covariate_value]]
            ).to(self.device).long().unsqueeze(1)
        
        embeddings = self.module.covars_embeddings[covariate](covar_ids).detach().cpu().numpy() # (n_covars, n_latent)
        
        covar_latent_adata = AnnData(X=embeddings)
        covar_latent_adata.obs[covariate] = [covariate_value] if covariate_value is not None else self.covars_encoder[covariate].keys()

        return covar_latent_adata

    def save(
        self,
        dir_path: str,
        overwrite: bool = False,
        save_anndata: bool = False,
        **anndata_write_kwargs,
    ):
        """
        Saves the state of the model.

        Parameters
        ----------
        dir_path : `str`
            Path to a directory.
        overwrite : `bool`, optional (default: `False`)
            Whether to overwrite the model/data in `dir_path` if it already exists.
        save_anndata : `bool`, optional (default: `False`)
            Whether to save the anndata along with the model.
        **anndata_write_kwargs : keyword arguments
            Keyword arguments to pass to anndata's write function.
        """
        os.makedirs(dir_path, exist_ok=True)

        # save public dictionaries
        total_dict = {
            "pert_encoder": self.pert_encoder,
            "covars_encoder": self.covars_encoder,
            "pert_smiles_map": self.pert_smiles_map,
        }

        json_dict = json.dumps(total_dict)
        with open(os.path.join(dir_path, "CPA_info.json"), "w") as f:
            f.write(json_dict)

        if isinstance(self.epoch_history, dict):
            self.epoch_history = pd.DataFrame().from_dict(
                self.training_plan.epoch_history
            )
            self.epoch_history.to_csv(
                os.path.join(dir_path, "history.csv"), index=False
            )
        elif isinstance(self.epoch_history, pd.DataFrame):
            self.epoch_history.to_csv(
                os.path.join(dir_path, "history.csv"), index=False
            )

        return super().save(
            dir_path=dir_path,
            overwrite=overwrite,
            save_anndata=save_anndata,
            **anndata_write_kwargs,
        )

    @classmethod
    def load(
        cls,
        dir_path: str,
        adata: Optional[AnnData] = None,
        use_gpu: Optional[Union[str, int, bool]] = None,
    ):
        """
        Loads the model from the specified directory.

        Parameters
        ----------
        dir_path : `str`
            Path to saved model.
        adata : `~anndata.AnnData`, optional (default: `None`)
            Annotated data matrix. Will call `cpa.CPA.setup_anndata` on the data after model restoration.
        use_gpu : `bool` or `str` or `int`, optional (default: `None`)
            Whether a GPU should be used. If `True`, will use GPU. 
        
        Returns
        -------
        :class:`~scvi.core.models.CPA`
            Restored model from the specified directory.
        """
        # load public dictionaries
        with open(os.path.join(dir_path, "CPA_info.json")) as f:
            total_dict = json.load(f)

            cls.pert_encoder = total_dict["pert_encoder"]
            cls.covars_encoder = total_dict["covars_encoder"]
            cls.pert_smiles_map = total_dict.get("pert_smiles_map", None)
            

        model = super().load(dir_path, adata, use_gpu)

        try:
            model.epoch_history = pd.read_csv(os.path.join(dir_path, "history.csv"))
        except:
            print("WARNING: The history was not found.")

        return model
